# $ANTLR 3.5.2 cadbiom/models/biosignal/translators/sigexpr_compiler.g 2018-12-14 16:41:05

import sys
from antlr3 import *
from antlr3.compat import set, frozenset

from antlr3.tree import *



from cadbiom.models.biosignal.sig_expr import *
import string



# for convenience in actions
HIDDEN = BaseRecognizer.HIDDEN

# token types
EOF=-1
AND=4
CHG=5
COM=6
COMMENT=7
CONSTR=8
DEF=9
DEFAULT=10
DIGIT=11
DOL=12
EG=13
EVENT=14
EXC=15
EXP=16
F=17
IDENT=18
INC=19
LETTER=20
MINUS=21
MUL=22
NOT=23
NOTEG=24
OR=25
PD=26
PG=27
PLUS=28
SCOL=29
SEQ=30
SYNC=31
T=32
WHEN=33
WS=34

# token names
tokenNames = [
    "<invalid>", "<EOR>", "<DOWN>", "<UP>",
    "AND", "CHG", "COM", "COMMENT", "CONSTR", "DEF", "DEFAULT", "DIGIT",
    "DOL", "EG", "EVENT", "EXC", "EXP", "F", "IDENT", "INC", "LETTER", "MINUS",
    "MUL", "NOT", "NOTEG", "OR", "PD", "PG", "PLUS", "SCOL", "SEQ", "SYNC",
    "T", "WHEN", "WS"
]




class sigexpr_compiler(Parser):
    grammarFileName = "cadbiom/models/biosignal/translators/sigexpr_compiler.g"
    api_version = 1
    tokenNames = tokenNames

    def __init__(self, input, state=None, *args, **kwargs):
        if state is None:
            state = RecognizerSharedState()

        super(sigexpr_compiler, self).__init__(input, state, *args, **kwargs)




        self.state_events = []
        self.free_clocks = []
        self.state_only = False
        self.state_only_in_bool = False
        self.catch_free_clocks = False
        self.deep = -1
        self.message = ""
        self.error_reporter = None


        self.delegates = []

        self._adaptor = None
        self.adaptor = CommonTreeAdaptor()



    def getTreeAdaptor(self):
        return self._adaptor

    def setTreeAdaptor(self, adaptor):
        self._adaptor = adaptor

    adaptor = property(getTreeAdaptor, setTreeAdaptor)


    def set_error_reporter(self, err):
        self.error_reporter = err

    def displayRecognitionError(self, tokenNames, re):
           hdr = self.getErrorHeader(re)
           msg = self.getErrorMessage(re, tokenNames)
           self.error_reporter.display('sig_exp ->'+hdr+msg+self.message)


    def displayExceptionMessage(self, e):
      msg = self.getErrorMessage(self, e, tokenNames)
      self.error_reporter.display('sig_exp ->'+msg)

    # semantic checks for compilers
    def check_ident(self, symbol_table, st_only, free_clock, deep, message, name):
        """
        Check if  name  is declared  (with state/input type if st_only = True)
        @return: a SigIdentExpr
        """
        try:
          name = name.encode("utf-8")
          type, s_deep = symbol_table[name]
          # for condition compiler
          if st_only:
            if (not (type == "state" or type=="input")):
              self.error_reporter.display("type error -> " + name + " is not a state (" +type+")"+ message)
          else:
            if deep >= 0:
              if s_deep >= deep:
                 self.error_reporter.display("type error -> " + name + " not declared in a surrounding macro state"+ message)
        except KeyError:
          if free_clock and not st_only:
            self.free_clocks.append(name)
          else:
            self.error_reporter.display("dec -> Undeclared event or state:" + name + message)
        return SigIdentExpr(name)

    def check_updown(self, symbol_table, id, mode):
        """
        This function introduce new signals: state> or state<
        """
        id  = id.encode("utf-8")
        try:
            att = symbol_table[id]
            type = att[0]
            s_deep = att[1]
        except KeyError:
            self.error_reporter.display('dec -> Undeclared state in variation:' + id)
            type = "error"
        if type == "state":
            if mode == 1: #up
                name = id+">"
            else: #down
                name = id+"<"
            self.state_events.append(name)
            return SigIdentExpr(name)
        else:
            self.error_reporter.display('type error -> Up and Down can only be derived from a state: '+id)

    def  check_change(self, symbol_table, id):
        id  = id.encode("utf-8")
        try:
            att = symbol_table[id]
            type = att[0]
            s_deep = att[1]
        except KeyError:
            self.error_reporter.display('dec -> Undeclared signal in variation:' + id)
            type = "error"
        if type == "state":
            refresh_expr = SigIdentExpr(id)
            st_expr = SigIdentExpr(id)
            return SigWhenExpr(SigConstExpr(True),SigDiffExpr(st_expr, refresh_expr))
        else:
            self.error_reporter.display('type error -> Change can only be derived  from a state: '+id)

    def check_sync(self, lexp):
        # type checking is done in expressions
        return SigConstraintExpr(SigConstraintExpr.SYNCHRO, lexp)

    def check_exclus(self, lexp):
        return SigConstraintExpr(SigConstraintExpr.EXCLU, lexp)


    def check_included(self, exp1,exp2):
        return SigConstraintExpr(SigConstraintExpr.INCL, [exp1,exp2])



    class sig_expression_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_expression_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_expression"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:163:1: sig_expression[tab_symb] returns [exp] : (exp1= sig_expression1[tab_symb] DOL |exp2= sig_constraint[tab_symb] DOL );
    def sig_expression(self, tab_symb):
        retval = self.sig_expression_return()
        retval.start = self.input.LT(1)


        root_0 = None

        DOL1 = None
        DOL2 = None
        exp1 = None
        exp2 = None

        DOL1_tree = None
        DOL2_tree = None

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:164:9: (exp1= sig_expression1[tab_symb] DOL |exp2= sig_constraint[tab_symb] DOL )
            alt1 = 2
            LA1_0 = self.input.LA(1)

            if (LA1_0 == DOL or LA1_0 == EVENT or (F <= LA1_0 <= IDENT) or LA1_0 == NOT or LA1_0 == PG or (T <= LA1_0 <= WHEN)) :
                alt1 = 1
            elif (LA1_0 == EXC or LA1_0 == INC or LA1_0 == SYNC) :
                alt1 = 2
            else:
                nvae = NoViableAltException("", 1, 0, self.input)

                raise nvae


            if alt1 == 1:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:164:11: exp1= sig_expression1[tab_symb] DOL
                root_0 = self._adaptor.nil()


                self._state.following.append(self.FOLLOW_sig_expression1_in_sig_expression148)
                exp1 = self.sig_expression1(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, exp1.tree)


                DOL1 = self.match(self.input, DOL, self.FOLLOW_DOL_in_sig_expression151)
                DOL1_tree = self._adaptor.createWithPayload(DOL1)
                self._adaptor.addChild(root_0, DOL1_tree)



                #action start
                retval.exp = ((exp1 is not None) and [exp1.exp] or [None])[0]
                #action end



            elif alt1 == 2:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:166:11: exp2= sig_constraint[tab_symb] DOL

                root_0 = self._adaptor.nil()


                self._state.following.append(self.FOLLOW_sig_constraint_in_sig_expression181)
                exp2 = self.sig_constraint(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, exp2.tree)


                DOL2 = self.match(self.input, DOL, self.FOLLOW_DOL_in_sig_expression184)
                DOL2_tree = self._adaptor.createWithPayload(DOL2)
                self._adaptor.addChild(root_0, DOL2_tree)



                #action start
                retval.exp = ((exp2 is not None) and [exp2.exp] or [None])[0]
                #action end



            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_expression"


    class sig_expression1_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_expression1_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_expression1"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:170:1: sig_expression1[tab_symb] returns [exp] : (exp1= sig_exp[tab_symb] ( DEFAULT ^exp2= sig_exp[tab_symb] )* |);
    def sig_expression1(self, tab_symb):
        retval = self.sig_expression1_return()
        retval.start = self.input.LT(1)


        root_0 = None

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:171:9: (exp1= sig_exp[tab_symb] ( DEFAULT ^exp2= sig_exp[tab_symb] )* |)
            alt3 = 2
            LA3_0 = self.input.LA(1)

            if (LA3_0 == EVENT or (F <= LA3_0 <= IDENT) or LA3_0 == NOT or LA3_0 == PG or (T <= LA3_0 <= WHEN)) :
                alt3 = 1
            elif (LA3_0 == COM or LA3_0 == DOL or LA3_0 == PD) :
                alt3 = 2
            else:
                nvae = NoViableAltException("", 3, 0, self.input)

                raise nvae


            if alt3 == 1:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:171:11: exp1= sig_exp[tab_symb] ( DEFAULT ^exp2= sig_exp[tab_symb] )*

                root_0 = self._adaptor.nil()


                self._state.following.append(self.FOLLOW_sig_exp_in_sig_expression1230)
                exp1 = self.sig_exp(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, exp1.tree)


                #action start
                retval.exp = ((exp1 is not None) and [exp1.exp] or [None])[0]
                #action end


                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:172:11: ( DEFAULT ^exp2= sig_exp[tab_symb] )*
                while True: #loop2
                    alt2 = 2
                    LA2_0 = self.input.LA(1)

                    if (LA2_0 == DEFAULT) :
                        alt2 = 1


                    if alt2 == 1:
                        # cadbiom/models/biosignal/translators/sigexpr_compiler.g:172:12: DEFAULT ^exp2= sig_exp[tab_symb]

                        DEFAULT3 = self.match(self.input, DEFAULT, self.FOLLOW_DEFAULT_in_sig_expression1247)
                        DEFAULT3_tree = self._adaptor.createWithPayload(DEFAULT3)
                        root_0 = self._adaptor.becomeRoot(DEFAULT3_tree, root_0)



                        self._state.following.append(self.FOLLOW_sig_exp_in_sig_expression1252)
                        exp2 = self.sig_exp(tab_symb)

                        self._state.following.pop()
                        self._adaptor.addChild(root_0, exp2.tree)


                        #action start
                        retval.exp = SigDefaultExpr(retval.exp, ((exp2 is not None) and [exp2.exp] or [None])[0])
                        #action end



                    else:
                        break #loop2



            elif alt3 == 2:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:176:11:

                root_0 = self._adaptor.nil()


                #action start
                retval.exp = SigConstExpr(True)
                #action end



            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_expression1"


    class sig_exp_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_exp_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_exp"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:179:1: sig_exp[tab_symb] returns [exp] : exp1= sig_bool[tab_symb] ( WHEN ^exp2= sig_bool[tab_symb] )* ;
    def sig_exp(self, tab_symb):
        retval = self.sig_exp_return()
        retval.start = self.input.LT(1)

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:180:9: (exp1= sig_bool[tab_symb] ( WHEN ^exp2= sig_bool[tab_symb] )* )
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:180:12: exp1= sig_bool[tab_symb] ( WHEN ^exp2= sig_bool[tab_symb] )*

            root_0 = self._adaptor.nil()


            self._state.following.append(self.FOLLOW_sig_bool_in_sig_exp338)
            exp1 = self.sig_bool(tab_symb)

            self._state.following.pop()
            self._adaptor.addChild(root_0, exp1.tree)


            #action start
            retval.exp = ((exp1 is not None) and [exp1.exp] or [None])[0]
            #action end


            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:181:11: ( WHEN ^exp2= sig_bool[tab_symb] )*
            while True: #loop4
                alt4 = 2
                LA4_0 = self.input.LA(1)

                if (LA4_0 == WHEN) :
                    alt4 = 1


                if alt4 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:181:12: WHEN ^exp2= sig_bool[tab_symb]

                    WHEN4 = self.match(self.input, WHEN, self.FOLLOW_WHEN_in_sig_exp355)
                    WHEN4_tree = self._adaptor.createWithPayload(WHEN4)
                    root_0 = self._adaptor.becomeRoot(WHEN4_tree, root_0)



                    #action start
                    st_only_save = self.state_only; self.state_only = self.state_only_in_bool
                    #action end


                    self._state.following.append(self.FOLLOW_sig_bool_in_sig_exp389)
                    exp2 = self.sig_bool(tab_symb)

                    self._state.following.pop()
                    self._adaptor.addChild(root_0, exp2.tree)


                    #action start
                    retval.exp = SigWhenExpr(retval.exp, ((exp2 is not None) and [exp2.exp] or [None])[0])
                    self.state_only = st_only_save
                    #action end



                else:
                    break #loop4




            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_exp"


    class sig_bool_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_bool_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_bool"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:190:1: sig_bool[tab_symb] returns [exp] : exp1= sig_bool_and[tab_symb] ( OR ^exp2= sig_bool_and[tab_symb] )* ;
    def sig_bool(self, tab_symb):
        retval = self.sig_bool_return()
        retval.start = self.input.LT(1)

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:191:9: (exp1= sig_bool_and[tab_symb] ( OR ^exp2= sig_bool_and[tab_symb] )* )
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:191:11: exp1= sig_bool_and[tab_symb] ( OR ^exp2= sig_bool_and[tab_symb] )*

            root_0 = self._adaptor.nil()


            self._state.following.append(self.FOLLOW_sig_bool_and_in_sig_bool471)
            exp1 = self.sig_bool_and(tab_symb)

            self._state.following.pop()
            self._adaptor.addChild(root_0, exp1.tree)


            #action start
            retval.exp = ((exp1 is not None) and [exp1.exp] or [None])[0]
            #action end


            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:192:11: ( OR ^exp2= sig_bool_and[tab_symb] )*
            while True: #loop5
                alt5 = 2
                LA5_0 = self.input.LA(1)

                if (LA5_0 == OR) :
                    alt5 = 1


                if alt5 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:192:12: OR ^exp2= sig_bool_and[tab_symb]

                    OR5 = self.match(self.input, OR, self.FOLLOW_OR_in_sig_bool487)
                    OR5_tree = self._adaptor.createWithPayload(OR5)
                    root_0 = self._adaptor.becomeRoot(OR5_tree, root_0)



                    self._state.following.append(self.FOLLOW_sig_bool_and_in_sig_bool492)
                    exp2 = self.sig_bool_and(tab_symb)

                    self._state.following.pop()
                    self._adaptor.addChild(root_0, exp2.tree)


                    #action start
                    retval.exp = SigSyncBinExpr("or", retval.exp, ((exp2 is not None) and [exp2.exp] or [None])[0])
                    #action end



                else:
                    break #loop5




            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_bool"


    class sig_bool_and_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_bool_and_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_bool_and"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:197:1: sig_bool_and[tab_symb] returns [exp] : exp1= sig_primary[tab_symb] ( AND ^exp2= sig_primary[tab_symb] )* ;
    def sig_bool_and(self, tab_symb):
        retval = self.sig_bool_and_return()
        retval.start = self.input.LT(1)

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:198:9: (exp1= sig_primary[tab_symb] ( AND ^exp2= sig_primary[tab_symb] )* )
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:198:11: exp1= sig_primary[tab_symb] ( AND ^exp2= sig_primary[tab_symb] )*

            root_0 = self._adaptor.nil()


            self._state.following.append(self.FOLLOW_sig_primary_in_sig_bool_and561)
            exp1 = self.sig_primary(tab_symb)

            self._state.following.pop()
            self._adaptor.addChild(root_0, exp1.tree)


            #action start
            retval.exp = ((exp1 is not None) and [exp1.exp] or [None])[0]
            #action end


            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:199:11: ( AND ^exp2= sig_primary[tab_symb] )*
            while True: #loop6
                alt6 = 2
                LA6_0 = self.input.LA(1)

                if (LA6_0 == AND) :
                    alt6 = 1


                if alt6 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:199:12: AND ^exp2= sig_primary[tab_symb]

                    AND6 = self.match(self.input, AND, self.FOLLOW_AND_in_sig_bool_and577)
                    AND6_tree = self._adaptor.createWithPayload(AND6)
                    root_0 = self._adaptor.becomeRoot(AND6_tree, root_0)



                    self._state.following.append(self.FOLLOW_sig_primary_in_sig_bool_and582)
                    exp2 = self.sig_primary(tab_symb)

                    self._state.following.pop()
                    self._adaptor.addChild(root_0, exp2.tree)


                    #action start
                    retval.exp = SigSyncBinExpr("and",retval.exp, ((exp2 is not None) and [exp2.exp] or [None])[0])
                    #action end



                else:
                    break #loop6




            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_bool_and"


    class sig_primary_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_primary_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_primary"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:204:1: sig_primary[tab_symb] returns [exp] : ( NOT ^nexp= sig_primary[tab_symb] |cexp= sig_constant | EVENT PG !exps= sig_expression1[tab_symb] PD !| WHEN PG !expw= sig_expression1[tab_symb] PD !|i7= IDENT UP |i8= IDENT DOWN |id2= IDENT | PG expse= sig_expression1[tab_symb] PD );
    def sig_primary(self, tab_symb):
        retval = self.sig_primary_return()
        retval.start = self.input.LT(1)

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:205:9: ( NOT ^nexp= sig_primary[tab_symb] |cexp= sig_constant | EVENT PG !exps= sig_expression1[tab_symb] PD !| WHEN PG !expw= sig_expression1[tab_symb] PD !|i7= IDENT UP |i8= IDENT DOWN |id2= IDENT | PG expse= sig_expression1[tab_symb] PD )
            alt7 = 8
            LA7 = self.input.LA(1)
            if LA7 == NOT:
                alt7 = 1
            elif LA7 == F or LA7 == T:
                alt7 = 2
            elif LA7 == EVENT:
                alt7 = 3
            elif LA7 == WHEN:
                alt7 = 4
            elif LA7 == IDENT:
                LA7 = self.input.LA(2)
                if LA7 == 3:
                    alt7 = 5
                elif LA7 == 2:
                    alt7 = 6
                elif LA7 == AND or LA7 == COM or LA7 == DEFAULT or LA7 == DOL or LA7 == OR or LA7 == PD or LA7 == WHEN:
                    alt7 = 7
                else:
                    nvae = NoViableAltException("", 7, 6, self.input)

                    raise nvae


            elif LA7 == PG:
                alt7 = 8
            else:
                nvae = NoViableAltException("", 7, 0, self.input)

                raise nvae


            if alt7 == 1:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:205:11: NOT ^nexp= sig_primary[tab_symb]

                root_0 = self._adaptor.nil()


                NOT7 = self.match(self.input, NOT, self.FOLLOW_NOT_in_sig_primary656)
                NOT7_tree = self._adaptor.createWithPayload(NOT7)
                root_0 = self._adaptor.becomeRoot(NOT7_tree, root_0)



                self._state.following.append(self.FOLLOW_sig_primary_in_sig_primary661)
                nexp = self.sig_primary(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, nexp.tree)


                #action start
                retval.exp = SigNotExpr(((nexp is not None) and [nexp.exp] or [None])[0])
                #action end



            elif alt7 == 2:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:208:11: cexp= sig_constant

                root_0 = self._adaptor.nil()


                self._state.following.append(self.FOLLOW_sig_constant_in_sig_primary704)
                cexp = self.sig_constant()

                self._state.following.pop()
                self._adaptor.addChild(root_0, cexp.tree)


                #action start
                retval.exp = ((cexp is not None) and [cexp.exp] or [None])[0]
                #action end



            elif alt7 == 3:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:211:11: EVENT PG !exps= sig_expression1[tab_symb] PD !

                root_0 = self._adaptor.nil()


                EVENT8 = self.match(self.input, EVENT, self.FOLLOW_EVENT_in_sig_primary751)
                EVENT8_tree = self._adaptor.createWithPayload(EVENT8)
                self._adaptor.addChild(root_0, EVENT8_tree)



                self.match(self.input, PG, self.FOLLOW_PG_in_sig_primary753)

                self._state.following.append(self.FOLLOW_sig_expression1_in_sig_primary758)
                exps = self.sig_expression1(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, exps.tree)


                self.match(self.input, PD, self.FOLLOW_PD_in_sig_primary761)

                #action start
                retval.exp = SigEventExpr(((exps is not None) and [exps.exp] or [None])[0])
                #action end



            elif alt7 == 4:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:214:11: WHEN PG !expw= sig_expression1[tab_symb] PD !

                root_0 = self._adaptor.nil()


                WHEN11 = self.match(self.input, WHEN, self.FOLLOW_WHEN_in_sig_primary798)
                WHEN11_tree = self._adaptor.createWithPayload(WHEN11)
                self._adaptor.addChild(root_0, WHEN11_tree)



                self.match(self.input, PG, self.FOLLOW_PG_in_sig_primary800)

                self._state.following.append(self.FOLLOW_sig_expression1_in_sig_primary805)
                expw = self.sig_expression1(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, expw.tree)


                self.match(self.input, PD, self.FOLLOW_PD_in_sig_primary808)

                #action start
                retval.exp = SigWhenExpr(SigConstExpr(True),((expw is not None) and [expw.exp] or [None])[0])
                #action end



            elif alt7 == 5:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:217:11: i7= IDENT UP

                root_0 = self._adaptor.nil()


                i7 = self.match(self.input, IDENT, self.FOLLOW_IDENT_in_sig_primary838)
                i7_tree = self._adaptor.createWithPayload(i7)
                self._adaptor.addChild(root_0, i7_tree)



                UP14 = self.match(self.input, 3, self.FOLLOW_3_in_sig_primary840)
                UP14_tree = self._adaptor.createWithPayload(UP14)
                self._adaptor.addChild(root_0, UP14_tree)



                #action start
                retval.exp = self.check_updown(tab_symb, i7.text, 1)
                #action end



            elif alt7 == 6:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:220:11: i8= IDENT DOWN

                root_0 = self._adaptor.nil()


                i8 = self.match(self.input, IDENT, self.FOLLOW_IDENT_in_sig_primary870)
                i8_tree = self._adaptor.createWithPayload(i8)
                self._adaptor.addChild(root_0, i8_tree)



                DOWN15 = self.match(self.input, 2, self.FOLLOW_2_in_sig_primary872)
                DOWN15_tree = self._adaptor.createWithPayload(DOWN15)
                self._adaptor.addChild(root_0, DOWN15_tree)



                #action start
                retval.exp = self.check_updown(tab_symb, i8.text, 2)
                #action end



            elif alt7 == 7:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:225:11: id2= IDENT

                root_0 = self._adaptor.nil()


                id2 = self.match(self.input, IDENT, self.FOLLOW_IDENT_in_sig_primary915)
                id2_tree = self._adaptor.createWithPayload(id2)
                self._adaptor.addChild(root_0, id2_tree)



                #action start
                retval.exp = self.check_ident(tab_symb, self.state_only, self.catch_free_clocks, self.deep, self.message, id2.text)
                #action end



            elif alt7 == 8:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:228:11: PG expse= sig_expression1[tab_symb] PD

                root_0 = self._adaptor.nil()


                PG16 = self.match(self.input, PG, self.FOLLOW_PG_in_sig_primary956)
                PG16_tree = self._adaptor.createWithPayload(PG16)
                self._adaptor.addChild(root_0, PG16_tree)



                self._state.following.append(self.FOLLOW_sig_expression1_in_sig_primary960)
                expse = self.sig_expression1(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, expse.tree)


                PD17 = self.match(self.input, PD, self.FOLLOW_PD_in_sig_primary963)
                PD17_tree = self._adaptor.createWithPayload(PD17)
                self._adaptor.addChild(root_0, PD17_tree)



                #action start
                retval.exp = ((expse is not None) and [expse.exp] or [None])[0]
                #action end



            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_primary"


    class sig_constant_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_constant_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_constant"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:232:1: sig_constant returns [exp] : ( T | F );
    def sig_constant(self, ):
        retval = self.sig_constant_return()
        retval.start = self.input.LT(1)


        root_0 = None

        T18 = None
        F19 = None

        T18_tree = None
        F19_tree = None

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:233:10: ( T | F )
            alt8 = 2
            LA8_0 = self.input.LA(1)

            if (LA8_0 == T) :
                alt8 = 1
            elif (LA8_0 == F) :
                alt8 = 2
            else:
                nvae = NoViableAltException("", 8, 0, self.input)

                raise nvae


            if alt8 == 1:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:233:12: T

                root_0 = self._adaptor.nil()


                T18 = self.match(self.input, T, self.FOLLOW_T_in_sig_constant1008)
                T18_tree = self._adaptor.createWithPayload(T18)
                self._adaptor.addChild(root_0, T18_tree)



                #action start
                retval.exp = SigConstExpr(True)
                #action end



            elif alt8 == 2:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:234:12: F

                root_0 = self._adaptor.nil()


                F19 = self.match(self.input, F, self.FOLLOW_F_in_sig_constant1023)
                F19_tree = self._adaptor.createWithPayload(F19)
                self._adaptor.addChild(root_0, F19_tree)



                #action start
                retval.exp = SigConstExpr(False)
                #action end



            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_constant"


    class sig_constraint_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.sig_constraint_return, self).__init__()

            self.exp = None
            self.tree = None





    # $ANTLR start "sig_constraint"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:237:1: sig_constraint[tab_symb] returns [exp] : ( SYNC ^ PG !el= exp_list[tab_symb] PD !| EXC ^ PG !el= exp_list[tab_symb] PD !| INC ^ PG !e3= sig_expression1[tab_symb] COM !e4= sig_exp[tab_symb] PD !);
    def sig_constraint(self, tab_symb):
        retval = self.sig_constraint_return()
        retval.start = self.input.LT(1)


        root_0 = None

        SYNC20 = None
        PG21 = None
        PD22 = None
        EXC23 = None
        PG24 = None
        PD25 = None
        INC26 = None
        PG27 = None
        COM28 = None
        PD29 = None
        el = None
        e3 = None
        e4 = None

        SYNC20_tree = None
        PG21_tree = None
        PD22_tree = None
        EXC23_tree = None
        PG24_tree = None
        PD25_tree = None
        INC26_tree = None
        PG27_tree = None
        COM28_tree = None
        PD29_tree = None

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:238:10: ( SYNC ^ PG !el= exp_list[tab_symb] PD !| EXC ^ PG !el= exp_list[tab_symb] PD !| INC ^ PG !e3= sig_expression1[tab_symb] COM !e4= sig_exp[tab_symb] PD !)
            alt9 = 3
            LA9 = self.input.LA(1)
            if LA9 == SYNC:
                alt9 = 1
            elif LA9 == EXC:
                alt9 = 2
            elif LA9 == INC:
                alt9 = 3
            else:
                nvae = NoViableAltException("", 9, 0, self.input)

                raise nvae


            if alt9 == 1:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:238:12: SYNC ^ PG !el= exp_list[tab_symb] PD !

                root_0 = self._adaptor.nil()


                SYNC20 = self.match(self.input, SYNC, self.FOLLOW_SYNC_in_sig_constraint1075)
                SYNC20_tree = self._adaptor.createWithPayload(SYNC20)
                root_0 = self._adaptor.becomeRoot(SYNC20_tree, root_0)



                PG21 = self.match(self.input, PG, self.FOLLOW_PG_in_sig_constraint1078)

                self._state.following.append(self.FOLLOW_exp_list_in_sig_constraint1083)
                el = self.exp_list(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, el.tree)


                PD22 = self.match(self.input, PD, self.FOLLOW_PD_in_sig_constraint1087)

                #action start
                retval.exp = self.check_sync(((el is not None) and [el.expl] or [None])[0])
                #action end



            elif alt9 == 2:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:241:12: EXC ^ PG !el= exp_list[tab_symb] PD !

                root_0 = self._adaptor.nil()


                EXC23 = self.match(self.input, EXC, self.FOLLOW_EXC_in_sig_constraint1138)
                EXC23_tree = self._adaptor.createWithPayload(EXC23)
                root_0 = self._adaptor.becomeRoot(EXC23_tree, root_0)



                PG24 = self.match(self.input, PG, self.FOLLOW_PG_in_sig_constraint1142)

                self._state.following.append(self.FOLLOW_exp_list_in_sig_constraint1149)
                el = self.exp_list(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, el.tree)


                PD25 = self.match(self.input, PD, self.FOLLOW_PD_in_sig_constraint1152)

                #action start
                retval.exp = self.check_exclus(((el is not None) and [el.expl] or [None])[0])
                #action end



            elif alt9 == 3:
                # cadbiom/models/biosignal/translators/sigexpr_compiler.g:244:12: INC ^ PG !e3= sig_expression1[tab_symb] COM !e4= sig_exp[tab_symb] PD !

                root_0 = self._adaptor.nil()


                INC26 = self.match(self.input, INC, self.FOLLOW_INC_in_sig_constraint1181)
                INC26_tree = self._adaptor.createWithPayload(INC26)
                root_0 = self._adaptor.becomeRoot(INC26_tree, root_0)



                PG27 = self.match(self.input, PG, self.FOLLOW_PG_in_sig_constraint1184)

                self._state.following.append(self.FOLLOW_sig_expression1_in_sig_constraint1189)
                e3 = self.sig_expression1(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, e3.tree)


                COM28 = self.match(self.input, COM, self.FOLLOW_COM_in_sig_constraint1192)

                self._state.following.append(self.FOLLOW_sig_exp_in_sig_constraint1197)
                e4 = self.sig_exp(tab_symb)

                self._state.following.pop()
                self._adaptor.addChild(root_0, e4.tree)


                PD29 = self.match(self.input, PD, self.FOLLOW_PD_in_sig_constraint1200)

                #action start
                retval.exp = self.check_included(((e3 is not None) and [e3.exp] or [None])[0], ((e4 is not None) and [e4.exp] or [None])[0])
                #action end



            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "sig_constraint"


    class exp_list_return(ParserRuleReturnScope):
        def __init__(self):
            super(sigexpr_compiler.exp_list_return, self).__init__()

            self.expl = None
            self.tree = None





    # $ANTLR start "exp_list"
    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:248:1: exp_list[tab_symb] returns [expl] : exp1= sig_expression1[tab_symb] ( COM exp2= sig_expression1[tab_symb] )* ;
    def exp_list(self, tab_symb):
        retval = self.exp_list_return()
        retval.start = self.input.LT(1)


        root_0 = None

        COM30 = None
        exp1 = None
        exp2 = None

        COM30_tree = None

        try:
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:249:10: (exp1= sig_expression1[tab_symb] ( COM exp2= sig_expression1[tab_symb] )* )
            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:249:12: exp1= sig_expression1[tab_symb] ( COM exp2= sig_expression1[tab_symb] )*

            root_0 = self._adaptor.nil()


            self._state.following.append(self.FOLLOW_sig_expression1_in_exp_list1257)
            exp1 = self.sig_expression1(tab_symb)

            self._state.following.pop()
            self._adaptor.addChild(root_0, exp1.tree)


            #action start
            retval.expl = [((exp1 is not None) and [exp1.exp] or [None])[0]]
            #action end


            # cadbiom/models/biosignal/translators/sigexpr_compiler.g:250:13: ( COM exp2= sig_expression1[tab_symb] )*
            while True: #loop10
                alt10 = 2
                LA10_0 = self.input.LA(1)

                if (LA10_0 == COM) :
                    alt10 = 1


                if alt10 == 1:
                    # cadbiom/models/biosignal/translators/sigexpr_compiler.g:250:14: COM exp2= sig_expression1[tab_symb]

                    COM30 = self.match(self.input, COM, self.FOLLOW_COM_in_exp_list1275)
                    COM30_tree = self._adaptor.createWithPayload(COM30)
                    self._adaptor.addChild(root_0, COM30_tree)



                    self._state.following.append(self.FOLLOW_sig_expression1_in_exp_list1279)
                    exp2 = self.sig_expression1(tab_symb)

                    self._state.following.pop()
                    self._adaptor.addChild(root_0, exp2.tree)


                    #action start
                    retval.expl.append(((exp2 is not None) and [exp2.exp] or [None])[0])
                    #action end



                else:
                    break #loop10




            retval.stop = self.input.LT(-1)


            retval.tree = self._adaptor.rulePostProcessing(root_0)
            self._adaptor.setTokenBoundaries(retval.tree, retval.start, retval.stop)



        except RecognitionException as re:
            self.reportError(re)
            self.recover(self.input, re)
            retval.tree = self._adaptor.errorNode(self.input, retval.start, self.input.LT(-1), re)

        return retval

    # $ANTLR end "exp_list"





    FOLLOW_sig_expression1_in_sig_expression148 = frozenset([12])
    FOLLOW_DOL_in_sig_expression151 = frozenset([1])
    FOLLOW_sig_constraint_in_sig_expression181 = frozenset([12])
    FOLLOW_DOL_in_sig_expression184 = frozenset([1])
    FOLLOW_sig_exp_in_sig_expression1230 = frozenset([1, 10])
    FOLLOW_DEFAULT_in_sig_expression1247 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_exp_in_sig_expression1252 = frozenset([1, 10])
    FOLLOW_sig_bool_in_sig_exp338 = frozenset([1, 33])
    FOLLOW_WHEN_in_sig_exp355 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_bool_in_sig_exp389 = frozenset([1, 33])
    FOLLOW_sig_bool_and_in_sig_bool471 = frozenset([1, 25])
    FOLLOW_OR_in_sig_bool487 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_bool_and_in_sig_bool492 = frozenset([1, 25])
    FOLLOW_sig_primary_in_sig_bool_and561 = frozenset([1, 4])
    FOLLOW_AND_in_sig_bool_and577 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_primary_in_sig_bool_and582 = frozenset([1, 4])
    FOLLOW_NOT_in_sig_primary656 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_primary_in_sig_primary661 = frozenset([1])
    FOLLOW_sig_constant_in_sig_primary704 = frozenset([1])
    FOLLOW_EVENT_in_sig_primary751 = frozenset([27])
    FOLLOW_PG_in_sig_primary753 = frozenset([14, 17, 18, 23, 26, 27, 32, 33])
    FOLLOW_sig_expression1_in_sig_primary758 = frozenset([26])
    FOLLOW_PD_in_sig_primary761 = frozenset([1])
    FOLLOW_WHEN_in_sig_primary798 = frozenset([27])
    FOLLOW_PG_in_sig_primary800 = frozenset([14, 17, 18, 23, 26, 27, 32, 33])
    FOLLOW_sig_expression1_in_sig_primary805 = frozenset([26])
    FOLLOW_PD_in_sig_primary808 = frozenset([1])
    FOLLOW_IDENT_in_sig_primary838 = frozenset([3])
    FOLLOW_3_in_sig_primary840 = frozenset([1])
    FOLLOW_IDENT_in_sig_primary870 = frozenset([2])
    FOLLOW_2_in_sig_primary872 = frozenset([1])
    FOLLOW_IDENT_in_sig_primary915 = frozenset([1])
    FOLLOW_PG_in_sig_primary956 = frozenset([14, 17, 18, 23, 26, 27, 32, 33])
    FOLLOW_sig_expression1_in_sig_primary960 = frozenset([26])
    FOLLOW_PD_in_sig_primary963 = frozenset([1])
    FOLLOW_T_in_sig_constant1008 = frozenset([1])
    FOLLOW_F_in_sig_constant1023 = frozenset([1])
    FOLLOW_SYNC_in_sig_constraint1075 = frozenset([27])
    FOLLOW_PG_in_sig_constraint1078 = frozenset([6, 14, 17, 18, 23, 27, 32, 33])
    FOLLOW_exp_list_in_sig_constraint1083 = frozenset([26])
    FOLLOW_PD_in_sig_constraint1087 = frozenset([1])
    FOLLOW_EXC_in_sig_constraint1138 = frozenset([27])
    FOLLOW_PG_in_sig_constraint1142 = frozenset([6, 14, 17, 18, 23, 27, 32, 33])
    FOLLOW_exp_list_in_sig_constraint1149 = frozenset([26])
    FOLLOW_PD_in_sig_constraint1152 = frozenset([1])
    FOLLOW_INC_in_sig_constraint1181 = frozenset([27])
    FOLLOW_PG_in_sig_constraint1184 = frozenset([6, 14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_expression1_in_sig_constraint1189 = frozenset([6])
    FOLLOW_COM_in_sig_constraint1192 = frozenset([14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_exp_in_sig_constraint1197 = frozenset([26])
    FOLLOW_PD_in_sig_constraint1200 = frozenset([1])
    FOLLOW_sig_expression1_in_exp_list1257 = frozenset([1, 6])
    FOLLOW_COM_in_exp_list1275 = frozenset([6, 14, 17, 18, 23, 27, 32, 33])
    FOLLOW_sig_expression1_in_exp_list1279 = frozenset([1, 6])



def main(argv, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    from antlr3.main import ParserMain
    main = ParserMain("sigexpr_lexer", sigexpr_compiler)

    main.stdin = stdin
    main.stdout = stdout
    main.stderr = stderr
    main.execute(argv)



if __name__ == '__main__':
    main(sys.argv)



## Filename    : simul_aux.py
## Author(s)   : Michel Le Borgne
## Created     : 04/2010
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA/INRIA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and INRIA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall INRIA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## INRIA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA/INRIA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux
##
"""
Auxiliary classes for simulator
"""

import sys
from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.guard_transitions.simulator.simul_exceptions import SimulException

class ChartExpVisitor(object):
    """
    Evaluator of event and condition expressions
    """
    def __init__(self, stable):
        self.symb_table = stable
        pass

    def visit_sig_ident(self, iex):
        """
        @param iex: ident expression
        """
        try:
            elem = self.symb_table[iex.name]
        except:
            print 'ChartExpVisitor:BIG BUG in expression evaluation '+iex.name
            sys.exit(1)
        return elem.activated

    def visit_sig_const(self, cex):
        """
        Simply value
        """
        return cex.value

    def visit_sig_default(self, dex):
        """
        Compute value
        """
        cv1 = dex.left_h.accept(self)
        cv2 = dex.right_h.accept(self)
        return cv1 or cv2

    def visit_sig_when(self, wex):
        """
        Compute value
        """
        cv1 = wex.left_h.accept(self)
        cv2 = wex.right_h.accept(self)
        return cv1 and cv2

    def visit_sig_equal(self, eex):
        """
        TODO
        """
        pass

    def visit_sig_diff(self, dex):
        """
        TODO
        """
        pass

    def visit_sig_not(self, nex):
        """
        Compute value
        """
        cval = nex.operand.accept(self)
        return not cval

    def visit_sig_event(self, eex):
        """
        TODO
        """
        pass

    def visit_sig_sync(self, sex):
        """
        compute value
        """
        cv1 = sex.left_h.accept(self)
        cv2 = sex.right_h.accept(self)
        if sex.operator == 'or':
            return cv1 or cv2
        if sex.operator == 'and':
            return cv1 and cv2


    def visit_sig_poly(self, pex):
        """
        TODO
        """
        pass

    def visit_sig_poly_pow(self, poe):
        """
        TODO
        """
        pass

class ModelExtractorVisitor(object):
    """
    Extract a submodel including places and transitions activated during a simulation.
    """
    def __init__(self, name_suffix, ex_type):
        """

        """
        self.name_suffix = name_suffix
        self.ex_type = ex_type
        pass

    def visit_chart_model(self, model):
        """
        Entrance point
        """
        self.sub_model = ChartModel(model.name+'_'+self.name_suffix)
        model.get_root().accept(self)
        return

    def visit_cstart_node(self, node):
        """
        Nothing to do
        """
        return

    def visit_ctrap_node(self, node):
        """
        Nothing to do
        """
        return

    def visit_csimple_node(self, node):
        """
        Extract nodes activated during simulation
        """
        if node.was_activated:
            self.sub_model.get_root().add_simple_node(node.name, 0.0, 0.0)
        return

    def visit_cinput_node(self, node):
        """
        Same as simple nodes
        """
        if node.was_activated:
            self.sub_model.get_root().add_input_node(node.name, 0.0, 0.0)
        return

    def visit_cperm_node(self, node):
        """
        nothing to do
        """
        return

    def visit_cmacro_node(self, node):
        """
        treated as simple nodes (no subnodes allowed)
        """
        if len(node.sub_nodes) > 0:
            mess = 'ModelExtractorVisitor: subnodes in macro state'
            mess = mess + 'are not allowed'
            raise SimulException()
        self.visit_csimple_node(node)
        return

    def visit_ctop_node(self, node):
        """
        visit subnodes then transitions
        """
        for snode in node.sub_nodes:
            snode.accept(self)
        for gtr in node.transitions:
            for trans in gtr:
                trans.accept(self)
        return

    def visit_ctransition(self, trans):
        """
        visit a transition - activated transition are extracted
        """
        if not trans.activated:
            return
        ori = trans.ori
        target = trans.ext
        # ori already extracted?
        try:
            n_ori = self.sub_model.get_node(ori.name)
        except: # add a copy to
            n_ori = self.sub_model.get_root().add_copy(ori)
        # same for target
        try:
            n_target = self.sub_model.get_node(target.name)
        except:
            n_target = self.sub_model.get_root().add_copy(target)
        # first time added or bug
        n_tr = self.sub_model.get_root().add_transition(n_ori, n_target)
        n_tr.set_event(trans.event)
        n_tr.set_condition(trans.condition)
        n_tr.activated = True # for path marks
        # add concurrent transitions not used
        if self.ex_type:
            for ctr in ori.outgoing_trans:
                if not trans == ctr and not ctr.activated:
                    ctarg = ctr.ext
                    try:
                        n_ctarg = self.sub_model.get_node(ctarg.name)
                    except:
                        n_ctarg = self.sub_model.get_root().add_copy(ctarg)
                    n_ctr = self.sub_model.get_root().add_transition(n_ori,
                                                                     n_ctarg)
                    if n_ctr: # may be already added here
                        n_ctr.set_event(ctr.event)
                        n_ctr.set_condition(ctr.condition)
                        n_ctr.activated = False # for path marks
        return




##
## Filename    : chart_lang.py
## Author(s)   : Michel Le Borgne
## Created     : 2/1/2012
## Revision    : 
## Source      : 
##
## Copyright 2009 - 2012 : IRISA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
##  have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
##     http:
##     mailto:
##
## Contributor(s): Geoffroy Andrieux
##


class LangVisitor(object):
    '''
    Decompiler from chart model to chart language description.
    '''
    PERM_N = '/p'
    INPUT_N = '/i'
    MACRO_N = '/m'
    
    MACRO = '/macro'
    ENDMACRO = '/endmacro'
    INDENT = "    "
    TARROW = " --> "
    
    def __init__(self, outfile):
        '''
        Constructor
        '''
        self.file = outfile
        self.indent = ""
        
    def visit_chart_model(self, mod):
        """
        Entry point in the model
        """
        if not mod.name == "":
            self.file.write("/name  "+ mod.name +"\n")
        mod.get_root().accept(self)
        if len(mod.constraints)>0: 
            self.file.write("/constraints\n")
            self.file.write(mod.constraints)
            self.file.write("\n/endconstraints\n")
        
    def visit_cstart_node(self, node):
        """
        Nothing to do
        """
        return ""
    
    def visit_ctrap_node(self, node):
        """
        Nothing to do
        """
        return ""
    
    def visit_csimple_node(self, node):
        """
        create an ident
        """
        return node.name
    
    def visit_cperm_node(self, node): 
        """
        return the mangled name
        """
        return node.name + self.PERM_N
    
    def visit_cinput_node(self, node): 
        """
        return the mangled name
        """
        return node.name + self.INPUT_N
    
    def visit_cmacro_node(self, mnode):
        """
        Generate the textual description of a macro
        """
        self.file.write("\n"+self.indent + self.MACRO + " "+ mnode.name + "\n")
        old_indent = self.indent
        self.indent = self.indent + self.INDENT
        # nodes
        for snode in mnode.sub_nodes:
            if not snode.is_macro(): 
                cond = (snode.incoming_trans == []) 
                cond = cond and (snode.outgoing_trans == [])
                if cond: # isolated nodes
                    dec = snode.accept(self)
                    if not dec == "":
                        self.file.write(self.indent + dec + " ;\n")
                # nodes used in transitions are not declared
            else:
                snode.accept(self)
        # transitions
        for trg in mnode.transitions:
            for trans in trg:
                trans.accept(self)
        # end
        self.indent = old_indent
        self.file.write(self.indent + self.ENDMACRO + "\n\n")
        return mnode.name
    
    def visit_ctop_node(self, tnode):    
        """
        Same as macro node except for indentation and constraints
        """
        # nodes
        self.file.write("// =============== macros nodes================\n\n")
        for snode in tnode.sub_nodes:
            if snode.is_macro():
                snode.accept(self)
        out_str = "// =============== isolated nodes ================\n\n"
        self.file.write(out_str)
        for snode in tnode.sub_nodes:
            if not snode.is_macro(): 
                cond = (snode.incoming_trans == []) 
                cond = cond and (snode.outgoing_trans == [])
                if cond: # isolated nodes
                    dec = snode.accept(self)
                    if not dec == "":
                        self.file.write(self.indent + dec + " ;\n")
        # transitions
        out_str = "\n\n// =============== transitions ================\n\n"
        self.file.write(out_str)
        for trg in tnode.transitions:
            for trans in trg:
                trans.accept(self)
        
        # constraints
        #TODO
        
    
    def visit_ctransition(self, trans): 
        """
        compile a transition
        """
        # macro nodes must be defined before use in transitions
        if not trans.ori.is_macro():   
            ori_txt = trans.ori.accept(self)
        else:
            ori_txt = trans.ori.name
        if not trans.ext.is_macro():
            target_txt = trans.ext.accept(self)
        else:
            target_txt = trans.ext.name
        event_expr = trans.event
        cond_expr = trans.condition
        action_expr = trans.action
        if event_expr and not trans.ori.is_start():
            guard = event_expr + "[" + cond_expr + "]"
        elif not trans.ori.is_start():
            guard = "[" + cond_expr + "]"
        else:
            guard = ""
        if action_expr:
            guard = guard + action_expr + " "
        else:
            guard = guard + " "
        out_str = self.indent + ori_txt + self.TARROW + target_txt + ";\n"
        self.file.write(out_str)
        self.file.write(self.indent + self.INDENT + guard)
        if not trans.note == "":
            note = "{" + trans.note + "}"
            self.file.write("\n"+ self.indent + self.INDENT + note)
        self.file.write(";\n\n")



# -*- coding: utf-8 -*-
## Filename    : TestStaticAnalysis.py
## Author(s)   : Michel Le Borgne
## Created     : 14 may. 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
from __future__ import print_function
import pkg_resources
import unittest
import sys
import os
import networkx as nx

from cadbiom.models.guard_transitions.translators.chart_xml_pid import PidXmlParser
from cadbiom.models.guard_transitions.analyser.ana_visitors import \
                                     DirectFlowGraphBuilder
from cadbiom.models.guard_transitions.analyser.static_analysis import \
                                     StaticAnalyzer
from cadbiom.models.guard_transitions.translators.chart_lang import LangVisitor
from cadbiom.models.guard_transitions.chart_model import ChartModel

from cadbiom.models.guard_transitions.translators.chart_xml import \
                                     MakeModelFromXmlFile
from cadbiom.commons import DIR_LOGS

TRACEFILE = sys.stdout
#TRACEFILE = open("/tmp/testMCLTranslator.txt",'w')


class Reporter(object):
    """
    Simple reporter for tests
    """
    def __init__(self):
        self.error = False
        self.mess = ""
        pass

    def display(self,  mes):
        """
        print message
        """
        self.error = True
        mess = self.mess + "ERROR -> "+mes
        TRACEFILE.write("\n\n"+ mess)

class TestAnaVisitors(unittest.TestCase):
    """
    Test visitor for static analysis
    """
    def test_frontier(self):
        """
        frontier visitor test
        """
        file_name = pkg_resources.resource_filename(
            __name__, # package name
            '../translators/tests/tgf_cano.xml'
        )
        reporter = Reporter()
        parser = PidXmlParser(file_name, reporter)
        if reporter.error:
            print('ERsROR IN READING FILE')

        cmodel = parser.model
        nx_graph = nx.DiGraph()
        gvi = DirectFlowGraphBuilder(nx_graph, False)
        cmodel.accept(gvi)

        aco = nx.attracting_components(nx_graph)
        scc = []
        for elem in aco:
            if len(elem)>1:
                scc.append(elem)
        print('SCC FRONTIER  ', scc)
        print(aco)



class TestStaticAnalysis(unittest.TestCase):
    """
    Test static analysis
    """
    def test_frontier1(self):
        """
        frontier computation
        """
        file_name = pkg_resources.resource_filename(
            __name__, # package name
            '../translators/tests/tgf_cano.xml'
        )
        reporter = Reporter()
        parser = PidXmlParser(file_name, reporter)
        if reporter.error:
            print('ERROR IN READING FILE')
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(parser.model)
        print('FRONTIER')
        ftr = sta.get_frontier_scc()
        print(ftr)
        # add a start transition on ft[0][1] (betaglycan__dimer___intToMb)
        assert 'betaglycan__dimer___intToMb' in ftr[0]
        parser.model.mark_as_frontier('betaglycan__dimer___intToMb')
        out = open(DIR_LOGS +
                   os.path.basename(os.path.splitext(file_name)[0]) +
                   '_cmp.cal',"w")
        # decompiler visitor
        lvi = LangVisitor(out)
        parser.model.accept(lvi)
        out.close()

    def test_frontier1_no_clock(self):
        """
        test with a rather big clockless model
        """
        file_name = pkg_resources.resource_filename(
            __name__, # package name
            '../translators/tests/tgf_cano.xml'
        )
        reporter = Reporter()
        parser = PidXmlParser(file_name, reporter, has_clock=False)
        if reporter.error:
            print('ERROR IN READING FILE')
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(parser.model)
        print('FRONTIER')
        ftr = sta.get_frontier_scc()
        print(ftr)
        # add a start transition on ft[0][1] (betaglycan__dimer___intToMb)
        assert 'betaglycan__dimer___intToMb' in ftr[0]
        parser.model.mark_as_frontier('betaglycan__dimer___intToMb')
        out = open(DIR_LOGS +
                   os.path.basename(os.path.splitext(file_name)[0]) +
                   '_noclock_cmp.cal',"w")
        # decompiler visitor
        lvi = LangVisitor(out)
        parser.model.accept(lvi)
        out.close()


    def test_frontier2(self):
        """
        frontier computation on all database
        """
        # test on all the database
        file_name = pkg_resources.resource_filename(
            __name__, # package name
            '../translators/tests/tgf_cano.xml'
        )
        reporter = Reporter()
        parser = PidXmlParser(file_name, reporter)
        if reporter.error:
            print('ERROR IN READING FILE')
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(parser.model)
        print('FRONTIER')
        print(sta.get_frontier_scc())

    def test_get_tr_node_variables(self):
        """
        n1 --> n2; h1 when n2 default h2 when n3[not n4]
        n1 --> n3; h2 when n4[]
        n4 --> n1; h3[]
        """
        model = ChartModel("Test")
        root = model.get_root()
        node1 = root.add_simple_node('n1', 0, 0)
        node2 = root.add_simple_node('n2', 0, 0)
        node3 = root.add_simple_node('n3', 0, 0)
        node4 = root.add_simple_node('n4', 0, 0)
        tr1 = root.add_transition(node1, node2)
        tr1.set_condition("not n4")
        tr1.set_event('h1 when n2 default h2 when n3')
        tr2 = root.add_transition(node1, node3)
        tr2.set_event('h2 when n4')
        tr3 = root.add_transition(node4, node1)
        tr3.set_event('h3')

        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(model)
        lnn, lhh = sta.test_get_tr_node_variables(tr1)
        print(lnn, lhh)

    def test_principal_variables(self):
        """
        n1 --> n2; h1 when n2 default h2 when n3[not n4]
        n1 --> n3; h2 when n4[n1 or (not n4)]
        n4 --> n1; h3[n3 or not n4]
        """
        model = ChartModel("Test")
        root = model.get_root()
        node1 = root.add_simple_node('n1', 0, 0)
        node2 = root.add_simple_node('n2', 0, 0)
        node3 = root.add_simple_node('n3', 0, 0)
        node4 = root.add_simple_node('n4', 0, 0)
        tr1 = root.add_transition(node1, node2)
        tr1.set_condition("not n4")
        tr1.set_event('h1 when n2 default h2 when n3')
        tr2 = root.add_transition(node1, node3)
        tr2.set_event('h2 when n4')
        tr2.set_condition('n1 or (not n4)')
        tr3 = root.add_transition(node4, node1)
        tr3.set_condition('n3 or (not n4)')
        tr3.set_event('h3')

        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(model)
        pva = sta.get_tr_principal_variables(tr1)
        print(pva)

    @unittest.skip("Test files not provided")
    def test_cadbiom_2_graphml(self):
        """
        test export in GML
        """
        file_name = '../translators/tests/pid0512.xml'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_pid_file(file_name, has_clock=True, ai_interpretation=0)
        mgraph = sta.make_dependence_dg()
        fname = '../translators/tests/pid0512_dependenceGraph.graphml'
        sta.export_2_graphml(mgraph, fname)

    @unittest.skip("Test files not provided")
    def test_get_pred_and_succ(self):
        """
        test predecessors and successors
        """
        file_name = '../../../bio_models/whole_pid_curated.bcx'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_file(file_name)
        graph = sta.make_dependence_dg()

        target = 'SNAIL'
        # PREDECESSORS
        pred = sta.get_predecessors(graph, target)
        print(target, ' has  ', len(pred), ' predecessors')

        # SUCCESSORS
        succ = sta.get_successors(graph, target)
        print(target, ' has  ', len(succ), ' successors')

    @unittest.skip("Test files not provided")
    def test_pred_dependence_graph(self):
        """
        As it says
        """
        file_name = '../../../bio_models/tgf/tgf_canonical.bcx'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_file(file_name)
        graph = sta.make_dependence_dg()

        target = 'p15INK4b'
        # PREDECESSORS
        pred = sta.get_predecessors(graph, target)
        graph = sta.remove_nodes_not_in_list(graph, pred)
        fname = '../translators/tests/tgf_cano_p15_pred.graphml'
        sta.export_2_graphml(graph, fname)

    @unittest.skip("Test files not provided")
    def test_pred_graph_extraction(self):
        """
        As it says
        """
        file_name = '../../../bio_models/tgf/tgf_canonical.bcx'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_file(file_name)
        graph = sta.make_dependence_dg()

        target = 'p15INK4b'
        # PREDECESSORS
        pred = sta.get_predecessors(graph, target)
        graph = sta.remove_nodes_not_in_list(graph, pred)

        #sta.extract_node_in_list(pred)

    @unittest.skip("Test files not provided")
    def test_gene_signal_dependent(self):
        """
        As it says
        """
        file_name = '/home/gandrieu/these/cadbiom model/fevrier2012_newParser/whole_pid_man_noInhib_modif.bcx'
        parser = MakeModelFromXmlFile(file_name)
        report = Reporter()
        model = parser.get_model()
        sta.build_from_chart_model(model)
        graph = sta.make_dependence_dg()
        # GENE RESEARCH
        gene_dict = dict() # key : gene, value : list of exCellRegion
        for node in graph.nodes():
            if len(node)>5:
                if node[-5:] == "_gene":
                    prot = node[:-5]
                    gene_dict[prot] = []
        # PREDECESSORS
        for gene in gene_dict.keys():
            gene_pred = sta.get_predecessors(graph, gene)
            for pred in gene_pred :
                if len(pred) > 13:
                    if pred[-13:] == "_exCellRegion":
                        gene_dict[gene].append(pred)
        print(gene_dict['p15INK4b'])
        print(gene_dict['p21CIP1'])
        # COUNTER
        dpt_dict = dict()
        for gene in gene_dict.keys():
            gene_dict[gene].sort()
            number_of_ex_cell_region = len(gene_dict[gene])
            if dpt_dict.has_key(number_of_ex_cell_region):
                dpt_dict[number_of_ex_cell_region] += 1
            else :
                dpt_dict[number_of_ex_cell_region] = 1
        dpt_items = dpt_dict.items()
        dpt_items.sort()
        for couple in dpt_items :
            print(couple[1], ' genes(s) that depends on ',
                  couple[0], ' signal(s)')


    @unittest.skip("Test files not provided")
    def test_signal_gene_influent(self):
        """
        ???
        """
        file_name = '/home/gandrieu/these/cadbiom model/fevrier2012_newParser/whole_pid_man_noInhib_modif.bcx'
        parser = MakeModelFromXmlFile(file_name)

        report = Reporter()
        model = parser.get_model()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(model)
        graph = sta.make_dependence_dg()
        # SIGNAL RESEARCH
        signal_dict = dict() # key : gene, value : list of exCellRegion
        for node in graph.nodes():
            if len(node)>13:
                if node[-13:] == "_exCellRegion":
                    signal_dict[node] = []
        # GENE RESEARCH
        gene_list = []
        for node in graph.nodes():
            if len(node)>5:
                if node[-5:] == "_gene":
                    prot = node[:-5]
                    gene_list.append(prot)
        # SUCCESSORS
        for signal in signal_dict.keys():
            signal_suc = sta.get_successors(graph, signal)
            for suc in signal_suc :
                if suc in gene_list :
                    signal_dict[signal].append(suc)
        # COUNTER
        dpt_dict = dict()
        for signal in signal_dict.keys():
            signal_dict[signal].sort()
            number_of_genes = len(signal_dict[signal])
            if dpt_dict.has_key(number_of_genes):
                dpt_dict[number_of_genes] += 1
            else :
                dpt_dict[number_of_genes] = 1
        dpt_items = dpt_dict.items()
        dpt_items.sort()
        for couple in dpt_items :
            print(couple[1],' signal(s) regulates ', couple[0],' gene(s)')

    @unittest.skip("Test files not provided")
    def test_ordering_pred(self):
        """
        write a predecessor file
        clustering predecessors into 10 class in fact of the distance of the shortest path to a given target
        Each line of the output file is a tab separated list of predecessors of the same class
        """

        file_name = '/home/gandrieu/these/models/cadbiom model/fevrier2012_newParser/whole_pid_man.bcx'
        parser = MakeModelFromXmlFile(file_name)

        report = Reporter()
        model = parser.get_model()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_model(model)
        graph = sta.make_dependence_dg()
        prop = 'SNAIL'
        pred_list = sta.get_predecessors(graph, prop)
        max_dist = 0
        for pred in pred_list:
            dist = nx.shortest_path_length(graph, source=pred, target=prop)
            if dist > max_dist :
                max_dist = dist


        ordering_pred_list = [[], [], [], [], [], [], [], [], [], []]
        for pred in pred_list :
            dist = nx.shortest_path_length(graph, source=pred, target=prop)
            new_dist = dist * 100 / max_dist

            if new_dist <= 10 :
                ordering_pred_list[0].append(pred)
            elif new_dist <= 20 :
                ordering_pred_list[1].append(pred)
            elif new_dist <= 30 :
                ordering_pred_list[2].append(pred)
            elif new_dist <= 40 :
                ordering_pred_list[3].append(pred)
            elif new_dist <= 50 :
                ordering_pred_list[4].append(pred)
            elif new_dist <= 60 :
                ordering_pred_list[5].append(pred)
            elif new_dist <= 70 :
                ordering_pred_list[6].append(pred)
            elif new_dist <= 80 :
                ordering_pred_list[7].append(pred)
            elif new_dist <= 90 :
                ordering_pred_list[8].append(pred)
            else :
                ordering_pred_list[9].append(pred)

        fileOut='/home/gandrieu/these/gephi/script/snailOrderingPred.txt'
        fih = open(fileOut,'w')
        for pred_list in ordering_pred_list:
            for pred in pred_list:
                fih.write(pred + '\t')
            fih.write('\n')
        fih.close()

    @unittest.skip("Test files not provided")
    def testStaticExtraction(self):
#        file_name = '../../../bio_models/whole_pid_man_noInhib_modif.bcx' # model without inhibitor
        file_name = '/home/gandrieu/these/processHitting/snail/whole_pid_man.bcx'
        parser = MakeModel_from_xml_file(file_name)

        report = Reporter()
        model = parser.get_model()
        sa = StaticAnalyzer(report)
        sta.build_from_chart_model(model)
        graph=sa.make_dependence_dg()

        # PREDECESSORS
#        pred = sa.get_predecessors(graph,'SNAIL')
#        pred=list(set(pred))
#        print(len(pred))

        # LOAD FILE
        pred =[]
        f = open('/home/gandrieu/these/processHitting/snail/whole_pid_man_SNAIL_processes_nameList.txt','r')
        for line in f :
            line=line[:-1]
            pred.append(line)
        f.close()
        pred = list(set(pred))

#        f=open('/home/gandrieu/tmp/snail_pred_all_dependences.txt','w')
#        for node in pred:
#            f.write(node)
#            f.write('\n')
#        f.close()

#        pred = sa.get_list_neighbors(graph, pred)
#        print(len(pred))
        # EXTRACTION
#        file_name = '../../../bio_models/whole_pid_curated.bcx'
#        parser = Make_model_from_xml_file(file_name)
        ex = SubModelExtraction(model,None,None)
        new_model = ex.extract_sub_model3(pred)
        new_model = perm_node_filter(new_model)
        xml = Xml_visitor(new_model)
        file = open("/home/gandrieu/tmp/whole_pid_man_extraction_SNAIL_PH_with_local_env.bcx",'w')
        file.write(xml.return_xml())
        file.close()

    @unittest.skip("Test files not provided")
    def test_statistics(self):
        file_name = '../../../bio_models/pid_or_clock.bcx'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_file(file_name)
        stat_file = open('../../../bio_models/stat.txt','w')
        sta.get_statistics(stat_file)

    @unittest.skip("Test files not provided")
    def test_model_extraction(self):
        file_name = '../../../bio_models/pid_or_clock.bcx'
        report = Reporter()
        sta = StaticAnalyzer(report)
        sta.build_from_chart_file(file_name)
#        graph = sta.make_dependence_dg()
        graph = sta.make_transition_dg()
        mac = ['ADAM10','HIF1A_cy','GammaSecretase_plasmaMb','ARNT_nucl','Jagged1_intToMb','SMAD2_3','SP1','IPAS','NOTCH1_golgiAp','MIZ_1','SHARP_RBPJ_CtIP_CtBP_LSD1_nucl','TGFB_TGFBR2_TGFBR1_DAB2_active_intToMb','SMAD4_nucl']
        prop = ['p15INK4b','p21CIP1']
#        sta.get_frontier_successors(graph, mac, prop, True)
        sta.get_propagated_places(mac, prop)

if __name__ == '__main__':

    import sys
    sys.setrecursionlimit(10000)

    suiteFew = unittest.TestSuite()
    suiteFew.addTest(TestStaticAnalysis('test_statistics'))
    unittest.TextTestRunner(verbosity=2).run(suiteFew)





# -*- coding: utf-8 -*-
## Filename    : MCLAnalyser.py
## Author(s)   : Michel Le Borgne
## Created     : 03/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux - IRISA/IRSET
##
"""
Main class to perform dynamic analysis of Cadbiom chart models.

Discrete event system simulation is a standard simulation scheme which consists
in the repetition of the following actions until some halting condition
is satisfied:
- Find the current events and inputs (including free events)
- Find the state variables subject to change
- Perform the state evolution

## Here you will find a global view of the process of dynamic analysis

mac_search:
    Wrapper for __mac_exhaustive_search()
    Return a generator of FrontierSolution objects.

__mac_exhaustive_search:
    Return a generator of FrontierSolution objects.

    Test if the given query is satisfiable according the given maximum number of steps.
    Get the minimum of steps that are mandatory to reach the final property in the query.

    This task is made with: `sq_is_satisfiable(query, max_user_step)`
    Which call internal functions of the unfolder (this is beyond the scope of this doc)

        - Initialisation: `self.unfolder.init_with_query(query)`
        - Ultimate test: `self.unfolder.squery_is_satisfied(max_step)`

    Adjust the attribute `steps_before_check` of the query, this is the number of
    useless shifts that are needed by the unfolder before finding a solution.

    Reload in the query a list of frontiers values that constitute each solution
    previously found.
    These values are in DIMACS format (i.e, numerical values). Since we doesn't
    want further similar solutions, each value has a negative sign
    (which which means that a value is prohibited).


    On banni les solutions précédemment trouvées. Une solution est apparentée à
    un ensemble de places frontières;
    Notons qu'il n'y a pas de contraintes sur les trajectoires ayant permis
    d'obtenir un ensemble de places frontières.

    Pour bannir les solutions 2 choix se présentent:
        - Soit les joindre par des "and" logiques et contraindre chaque solution
        par un "non".
        Ex: Pour 2 solutions [(A, B), (B, C)]:
        not((A and B) or (B and C))

        - Soit conserver les valeurs de chaque ensemble de places frontière sous
        forme de clauses constituées de valeurs numériques.
        Ex: Pour 2 solutions [(A, B), (B, C)]:
        [[-1, -2], [-2, -3]]

    La deuxième solution est nettement plus performante car elle permet de
    s'affranchir du travail de parsing réalisé par l'intervention d'une grammaire,
    des propriétés de plus en plus complexes au fur et à mesure des solutions
    trouvées.

    Ainsi, chaque nouvelle requête se voit recevoir dans son attribut dim_start,
    l'ensemble des solutions précédentes, bannies, au format DIMACS.


    On cherche ensuite d'abord des solutions non minimales (actuellement 2) via:
    `self.__sq_dimacs_frontier_solutions(query, nb_step, 2)`

    Pour ensuite les élaguer en supprimant les places non essentielles à la
    satisfiabilité de la propriété via:
    `small_fsol = self.less_activated_solution(lfsol)`
    `current_mac = self.__prune_frontier_solution(small_sol, query, nb_step)`

    Ce processus itératif est le plus "time-consuming" car nous n'avons pas
    le controle sur les solutions fournies par SAT et les solutions non minimales
    sont en général très éloignées de la solution minimale, c.-à-d. contiennent
    beaucoup plus de places (ces places excédentaires sont dispensables).


next_mac(self, query, max_step):
    Search a Minimal Access Condition for the given query.

    This is a function very similar to `__mac_exhaustive_search()`, but it
    returns only 1 solution.
    Satisfiability tests and the banishment of previous solutions must be done
    before the call.

---

## Back on the few mentioned functions and their call dependencies:

less_activated_solution(self, dimacs_solution_list):
    Get the solution with the less number of activated frontier places among the given solutions.

__prune_frontier_solution(self, fsol, query, max_step):
    Take a frontier condition which induces a property "prop" and try to
    reduce the number of activated frontier variables from this solution while
    inducing "prop".

    Find at most `nb_sols_to_be_pruned` frontier solutions inducing
    the same final property but with all inactivated frontier places
    forced to be initially False.

    Repeat the operation until there is only one solution.

    Return <DimacsFrontierSol>

    This functions calls __solve_with_inact_fsolution().

__solve_with_inact_fsolution(self, dimacs_front_sol, query, max_step, max_sol):
    Frontier states not activated in dimacs_front_sol (DimacsFrontierSol)
    are **forced to be False at start**; solve this and return DimacsFrontierSol
    objects matching this constraint (their number is defined with the
    argument `max_sols`)
    Return <tuple <DimacsFrontierSol>>

    This function calls __sq_dimacs_frontier_solutions().

__sq_dimacs_frontier_solutions(self, query, max_step, max_sol)
    Search set of minimal solutions for the given query, with a maximum of steps
    Tuple of unique DimacsFrontierSol objects built from RawSolutions
    from the solver.
    Return <tuple <DimacsFrontierSol>>

    This function calls sq_solutions().

sq_solutions(query, max_step, max_sol, vvars)
    This function is the lowest level function exploiting the solver results.
    Return <list <RawSolution>>

    This function calls:
        self.unfolder.init_with_query(query)
        self.unfolder.squery_solve(vvars, max_step, max_sol)


## Quiet deprecated but still used in the GUI:

sq_frontier_solutions(self, query, max_step, max_sol):
    This function is a wrapper of sq_solutions().
    Return <list <FrontierSolution>>
"""
from __future__ import print_function

from antlr3 import ANTLRFileStream, CommonTokenStream

from cadbiom.models.guard_transitions.translators.chart_xml import \
        MakeModelFromXmlFile

from cadbiom.models.guard_transitions.translators.chart_xml_pid import \
        MakeModelFromPidFile

from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.clause_constraints.mcl.MCLTranslators import  GT2Clauses
from cadbiom.models.clause_constraints.mcl.CLUnfolder import CLUnfolder
from cadbiom.models.clause_constraints.mcl.MCLSolutions import MCLException

from cadbiom.models.clause_constraints.CLDynSys import CLDynSys
from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser
from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.clause_constraints.mcl.MCLSolutions import FrontierSolution, DimacsFrontierSol
from cadbiom.models.clause_constraints.mcl.MCLQuery import MCLSimpleQuery
from cadbiom import commons as cm

# Standard imports
from logging import DEBUG, INFO

LOGGER = cm.logger()


class MCLAnalyser(object):
    """Query a UCL dynamical system and analyse solutions

    When loading a model in a MCLAnalyser object, a CLUnfolder is generated which
    implies the compilation of the dynamical system into a clause constraint
    dynamical system and the DIMACS coding of all the variables.
    This coding cannot be changed later.
    The unfolding of constraints is performed efficiently on the numeric form
    of the clause constraints. The CLUnfolder provide various method to convert
    variable names to DIMACS code and back, to extract the frontier of a model
    and to extract informations from raw solutions.

    Attributes and default values:

        :param dynamical_system: dynamical system in clause constraint form
        :param unfolder: computation management: unfolding
        :param reporter: reporter for generic error display
        :param translator_opti: turn on optimizations for ANTLR translation
            (subexpression elimination)
        :param nb_sols_to_be_pruned: For mac search: We search a number of
            solutions that will be pruned later, in order to find the most
            optimized MAC with a reduced the number of activated frontiers.
        :key debug: (optional) Used to activate debug mode in the Unfolder

        :type dynamical_system: None before loading a model / <CLDynSys>
        :type unfolder: None before loading a model / <CLUnfolder>
        :type debug: False / <boolean>
        :type reporter: <ErrorRep>
        :type translator_opti: True / <boolean>
        :type nb_sols_to_be_pruned: 10
    """

    def __init__(self, report, debug=False):
        """The built analyser is void by default

        @param report: a standard reporter
        """
        self.debug = debug
        self.unfolder = None
        self.reporter = report
        self.translator_opti = True

        # For mac search: We search a number of solutions that will be pruned
        # in order to find the most optimized MAC with a reduced the number of
        # activated frontiers.
        ## TODO: ajuster ce paramètre dynamiquement
        self.nb_sols_to_be_pruned = 10

    @property
    def dynamical_system(self):
        """Return the CLDynSys object of the current unfolder"""
        return self.unfolder.dynamical_system

    ## Building MCLAnalyser ####################################################

    def build_from_chart_model(self, model):
        """Build an MCLAnalyser and its CLUnfolder from a chartmodel object.

        Every parser uses this function at the end.

        @param model: <ChartModel>
        """
        # Reloading a MCLAnalyser is forbidden - create a new one is mandatory
        if self.unfolder:
            raise MCLException("This MCLAnalyser is already initialized")

        # Build CLDynSys (dynamical system in clause constraint form)
        # Parse the model data
        vtab = TableVisitor(self.reporter)
        model.accept(vtab)
        if self.reporter.error:
            return
        # PS:
        # Here, all keys of vtab.tab_symb = no_frontiers + frontiers + model name
        # (probably a bug for this last one)
        dynamic_system = CLDynSys(vtab.tab_symb, self.reporter)
        if self.reporter.error:
            return
        # Build clauses and auxiliary variables from events and places in the model
        comp_visitor = GT2Clauses(dynamic_system, self.reporter, self.translator_opti)
        model.accept(comp_visitor)
        if self.reporter.error:
            return
        # Build unfolder
        self.unfolder = CLUnfolder(dynamic_system, debug=self.debug)

    def build_from_chart_file(self, file_name):
        """Build an MCLAnalyser from a .bcx file

        @param file_name: str - path of the .bcx file
        """
        parsing = MakeModelFromXmlFile(file_name)
        # chart model
        self.build_from_chart_model(parsing.get_model())

    def build_from_pid_file(self, file_name, has_clock=True, ai_interpretation=0):
        """Build an MCLAnalyser from a .xml file of PID database

        @param file_name: str - path of .xml file
        """
        parsing = MakeModelFromPidFile(
            file_name, self.reporter, has_clock, ai_interpretation
        )
        # chart model
        self.build_from_chart_model(parsing.get_model())

    def build_from_cadlang(self, file_name):
        """Build an MCLAnalyser from a .cal, Cadlang file

        @param file_name: str - path of .cal file
        """
        creporter = self.reporter
        fstream = ANTLRFileStream(file_name)
        lexer = cadlangLexer(fstream)
        lexer.set_error_reporter(creporter)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(creporter)
        model = ChartModel(file_name)
        parser.cad_model(model)
        # chart model
        self.build_from_chart_model(parser.model)

    ## Solutions processing ####################################################

    def less_activated_solution(self, dimacs_solution_list):
        """Get the solution with the less number of activated frontier places
        among the given solutions.

        ..note:: There is not constraint on the complexity of the trajectories;
            only on the number of activated frontier places.

        :param dimacs_solution_list: list of DimacsFrontierSol objects
        :return: 1 DimacsFrontierSol with the less number of activated states.
        :type dimacs_solution_list: <tuple <DimacsFrontierSol>>
        :rtype: <DimacsFrontierSol>
        """
        # Count the number of activated places in each solution
        activated_per_sol = [sol.nb_activated_frontiers for sol in dimacs_solution_list]
        return dimacs_solution_list[activated_per_sol.index(min(activated_per_sol))]


    def __solve_with_inact_fsolution(self, dimacs_front_sol, query, max_step, max_sol):
        """Frontier states not activated in dimacs_front_sol (DimacsFrontierSol)
        are **forced to be False at start**; solve this and return DimacsFrontierSol
        objects matching this constraint (their number is defined with the
        argument `max_sol`).

        :param dimacs_front_sol: A solution to obtain the property.
        :param query: Current query
        :param max_step: Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        :param max_sol: Number of wanted solutions.
        :return: DimacsFrontierSol objects matching the constraint of
            useless inactivated frontier places.
        :type dimacs_front_sol: <DimacsFrontierSol>
        :type query: <MCLSimpleQuery>
        :type max_step: <int>
        :type max_sol: <int>
        :return: <tuple <DimacsFrontierSol>>
        :precondition: dimacs_front_sol is a frontier solution for the query
        """
        # Collect inactivated frontier places
        # OLD:
        # inactive_frontier_values = \
        #     [[var] for var in dimacs_front_sol.frontier_values if var < 0]
        # Intersection between pre-computed negative values for all frontiers
        # (negative and positive) in the given solution, and all frontiers.
        inactive_frontier_values = \
            [[var] for var in
             self.unfolder.frontiers_negative_values & dimacs_front_sol.frontier_values]

        # Append inactivated frontier places as a start property in DIMACS form
        # PS: do not test dim_start: In the worse case, it's an empty list.
        query.dim_start += inactive_frontier_values

        # Must return fsol even if there is no more solution (search first 10 sols)
        # Get <tuple <DimacsFrontierSol>>
        return self.__sq_dimacs_frontier_solutions(query, max_step, max_sol)


    def __prune_frontier_solution(self, fsol, query, max_step):
        """"Take a frontier condition which induces a property "prop" and try to
        reduce the number of activated frontier variables from this solution while
        inducing "prop".

        Find at most `nb_sols_to_be_pruned` frontier solutions inducing
        the same final property but with all inactivated frontier places
        forced to be initially False.

        Repeat the operation until there is only one solution.

        :type fsol: <DimacsFrontierSol>
        :type query: <MCLSimpleQuery>
        :rtype: <DimacsFrontierSol>

        ASSERT: fsol is a frontier condition implying sq satisfiability
        """
        # for debug - too expansive to be activated anytime
        # assert(self.sq_is_satisfiable(query, max_step))
        sol_to_prune = fsol
        # Number of activated places in solution
        # We are trying to reduce this number
        current_len = fsol.nb_activated_frontiers
        next_len = 0
        is_pruned = True

        i = 0
        while is_pruned:
            i += 1
            # find at most nb_sols_to_be_pruned frontier solutions inducing
            # the same final property but with all inactivated frontier places
            # forced to be initially False.
            # return sol_to_prune if no more solution

            # PS: We need to index the result in less_activated_solution
            # thus, this why __solve_with_inact_fsolution() returns a tuple.
            dimacs_solutions = \
                self.__solve_with_inact_fsolution(
                    sol_to_prune,
                    query,
                    max_step,
                    self.nb_sols_to_be_pruned
                )

            LOGGER.info(
                "MCLA::Prune: Iteration %s, nb pruned inact solutions: %s",
                i,
                len(dimacs_solutions)
            )
            ####TODO: regarder si ces solutions sont si différentes et ce qu'on en fait au final...
            ### en fin de compte il est possible que le solveur ne soit pas si bete...
            ##TODO:  ajuster automatiquement nb_sols_to_be_pruned

            # PS: dimacs_solutions should never be None or empty
            if len(dimacs_solutions) == 1:
                # No new solution has been found,
                # the current solution is the best (sol_to_prune)
                return dimacs_solutions[0] # it's sol_to_prune variable

            # Seach for less activated solution
            # Get the solution with the less number of activated states
            next_sol = self.less_activated_solution(dimacs_solutions)
            next_len = next_sol.nb_activated_frontiers
            is_pruned = current_len > next_len

            LOGGER.info(
                "MCLA::Prune: Current solution activated length:%s next:%s",
                current_len, next_len
            )

            if is_pruned:
                sol_to_prune = next_sol
                current_len = next_len

        # never reach this point
        raise MCLException("MCLAnalyser:__prune_frontier_solution:: what happened?")

    ## Generic solver for simple queries #######################################

    def sq_is_satisfiable(self, query, max_step):
        """Check if the query is satisfiable.

        ..warning::
            Many logical variables are auxiliary variables with no interest other
            than technical, we want different solutions to differ by their values
            on some set of significant variables.

            **These variables of interest are not checked here!**
                It is just a satisfiability test!

        ..warning:: No frontier place is initialized to False in simple queries.

        :param query:
        :param max_step: Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        :type query: <MCLSimpleQuery>
        :type max_step: <int>
        :rtype: <boolean>
        """
        # Initialise the unfolder with the given query
        # Following properties are used from the query:
        #    start_prop, dim_start, inv_prop, dim_inv, final_prop, dim_final,
        #    variant_prop, dim_variant_prop, steps_before_check
        self.unfolder.init_with_query(query)
        # go
        return self.unfolder.squery_is_satisfied(max_step)


    def sq_solutions(self, query, max_step, max_sol, vvars):
        """Return a list of RawSolution objects

        This function is the lowest level function exploiting the solver results.

        Parameters are the same as in sq_is_satisfiable() except for vvars
        parameter which deserves some explanations.

        The solver doesn’t provide all solutions of the set of logical constraints
        encoding the temporal property. It gives only a sample of these solutions
        limited in number by the max_sol parameter. **Since many logical variables
        are auxiliary variables with no interest other than technical, we want
        different solutions to differ by their values on some set of significant
        variables. That is the meaning of the vvars parameter.**

        Remark that the variables of interest must be specified in DIMACS code.
        Most of the time, this method is used internally, so DIMACS code is
        transparent to the user.

        The class RawSolution provides a representation of solutions which are
        raw results from the SAT solver with all variable parameters from the
        unfolder, and consequently not easily readable.
        In addition to the solution in DIMACS form (including values for auxiliary
        variables), a RawSolution object registers data which permit information
        extraction from these raw solutions. The most important methods are:

            - frontier_pos_and_neg_values:
                Which extracts the frontier values from the solution.
                These values are in DIMACS code.

            - extract_activated_frontier_values():
                Extracts frontier variables which are active in solution.

            - extract_act_input_clock_seq():
                Extract the sequence of activated inputs and events in the solution.


        In models of signal propagation, we are more interested in frontier solutions.
        The following method is then well adapted: sq_frontier_solutions(),
        because it returns FrontierSolutions.


        @param query: MCLSimpleQuery
        @param max_step: int - Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        :param vvars: Variables for which solutions must differ.
            In practice, it is a list with the values (Dimacs code)
            of all frontier places of the system.
        :type vvars: <list <int>>
        :return: a list of RawSolutions from the solver
        :rtype: <list <RawSolution>>

        @warning:  no_frontier places are initialized to False in simple queries
            except if initial condition.

        @warning: if the query includes variant constraints, the horizon (max_step) is
            automatically adjusted to min(max_step, len(variant_constraints)).
        """
        # Initialise the unfolder with the given query
        # Following properties are used from the query:
        #    start_prop, dim_start, inv_prop, dim_inv, final_prop, dim_final,
        #    variant_prop, dim_variant_prop, steps_before_check
        self.unfolder.init_with_query(query)
        # go
        return self.unfolder.squery_solve(vvars, max_step, max_sol)


    def sq_frontier_solutions(self, query, max_step, max_sol):
        """Compute active frontier places and timings

        Compute the set of frontier place names which must be activated for
        implying query satisfaction and returned it as a list of FrontierSolution.

        Quiet deprecated but referenced by:
            gt_gui/chart_checker/chart_checker_controler
            cadbiom_cmd/solution_search.py very old code for debug purpose
            self.mac_inhibitor_search

        In models of signal propagation, we are more interested in frontier solutions.
        So the current method is better adapted: sq_frontier_solutions()
        than sq_solutions() which returns RawSolutions.

        This function is a wrapper of sq_solutions().


        The class FrontierSolution provides objects wrapping a symbolic
        representation of frontier values and timing from a raw solution.
            - The main attributes are activated_frontier which is a set of
            (string) names of frontier places which are activated in the solution.

            - The second important attribute is ic_sequence, a list of strings
            which describes the successive activated inputs and free events in
            the solution.

        If events h2, h5 and input in3 are activated at step k in the solution,
        then the kth element of the list is “% h2 h5 in3”.
        This format is understood by the Cadbiom simulator.

        @warning: no frontier places are initialized to False

        @param query: MCLSimpleQuery
        @param max_step: int - Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        @param max_sol: max number of wanted solutions.
        @return: List of FrontierSolutions built from RawSolutions from the solver.
            So, lists of lists of frontier place names which must be activated
            for implying query satisfaction
        :rtype: <list <FrontierSolution>>
        """
        vvars = self.unfolder.frontier_values
        # sq_solutions() returns <list <RawSolution>>
        # We transtype them to FrontierSolution
        return [FrontierSolution.from_raw(sol) for sol
                in self.sq_solutions(query, max_step, max_sol, vvars)]


    def __sq_dimacs_frontier_solutions(self, query, max_step, max_sol):
        """Search set of minimal solutions for the given query, with a maximum of steps

        DimacsFrontierSol objects provides some interesting attributes like:
        activated_frontier_values (raw values of literals).
        Note: FrontierSolution are of higher level without such attributes.

        @param query: MCLSimpleQuery
        @param max_step: int - Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        @param max_sol: int - max number of solutions for each solve
        :return: Tuple of unique DimacsFrontierSol objects built from RawSolutions
            from the solver.
            .. note:: Unicity is based on the set of activated frontier values.
        :rtype: <tuple <DimacsFrontierSol>>
        """
        vvars = self.unfolder.frontier_values
        # sq_solutions() returns <list <RawSolution>>
        raw_sols = self.sq_solutions(query, max_step, max_sol, vvars)
        # Get unique DimacsFrontierSol objects from RawSolution objects
        dimacfrontsol = DimacsFrontierSol.get_unique_dimacs_front_sols_from_raw_sols(raw_sols)

        if LOGGER.getEffectiveLevel() == INFO:
            LOGGER.info(
                "MCLA: Nb queried: %s, steps: %s\n"
                "\tget: %s, unique: %s\n"
                "\tactivated fronts per dimacfrontsol: %s",
                max_sol, max_step,
                len(raw_sols), len(dimacfrontsol),
                [len(sol) for sol in dimacfrontsol]
            )
        return dimacfrontsol


    def __mac_exhaustive_search(self, query, max_user_step):
        """Return a generator of FrontierSolution objects.

        This is a function very similar to next_mac(), but it returns
        many solutions.
        The satisfiability tests and the banishment of the previous solutions
        are done internally in an optimal way.
        There is also a self adjustment of the number of minimum steps during
        which to find solutions until reaching the maximum number set by the user.

        ---

        On cherche ensuite d'abord des solutions non minimales (actuellement 2) via:
        self.__sq_dimacs_frontier_solutions(query, nb_step, 2)

        Pour ensuite les élaguer en supprimant les places non essentielles à la
        satisfiabilité de la propriété via:
        small_fsol = self.less_activated_solution(lfsol)
        current_mac = self.__prune_frontier_solution(small_sol, query, nb_step)

        Ce processus itératif est le plus "time-consuming" car nous n'avons pas
        le controle sur les solutions fournies par SAT et les solutions non minimales
        sont en général très éloignées de la solution minimale, c.-à-d. contiennent
        beaucoup plus de places (ces places excédentaires sont dispensables).

        :param query: Current query.
        :param max_user_step: Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        :type query: <MCLSimpleQuery>
        :type max_user_step: <int>
        :return: A generator of FrontierSolution objects which are a wrapper of
            a symbolic representation of frontier values and timings from
            a raw solution.
        :rtype: <generator <FrontierSolution>>
        """
        # list of timed minimal activation conditions on frontier (dimacs code)
        # i.e list<DimacsFrontierSol>
        # mac_list = [] # Not used anymore

        # Keep a list of frontier values to be banned in the next search
        forbidden_frontier_values = query.dim_start

        # Satisfiability test
        reachable = self.sq_is_satisfiable(query, max_user_step)

        # Get the minimum number of steps for reachability
        min_step = self.unfolder.get_current_step()
        # Adjust the number of useless shifts needed before finding a solution
        # (i.e. testing final prop).
        # __steps_before_check can't be >= max_user_step
        # (see squery_is_satisfied() and squery_solve())
        query.steps_before_check = min_step - 1

        LOGGER.info("Solution reachable in min_step: %s", min_step)

        while reachable:
            # Forbidden solutions: already discovered macs (see at the end of the while loop)
            # Ban all sets of activated frontiers (1 set = 1 solution)
            # Equivalent of wrapping each previous solution with a logical OR:
            # OR( not (frontier places))
            ## TODO: check in unfolder: dim_start elements joined by OR or AND operator ?
            ## TODO: keep previous dim_start à la première itération (mettre le contenu dans forbidden values)
            query.dim_start = list(forbidden_frontier_values)
            print("query variant prop", query.variant_prop)
            ## TODO: equivaut à mettre à jour le start_prop ou __initial_property avec l'ancienne méthode ??
            ## si non, faudrait ptetre éviter de recompiler ça à chaque fois dans __init_initial_constraint_0
            #### redondance code next_mac()
            # Solutions differ on frontiers: Search only 2 different solutions
            # to avoid all activated solutions
            # Get <tuple <DimacsFrontierSol>>
            dimacs_front_sol = self.__sq_dimacs_frontier_solutions(query, min_step, 2)
            if not dimacs_front_sol:
                # If we do not want to self-adjust steps, just break the loop
                # break

                LOGGER.info("MCLA: NOT FOUND -> satis test; current step %s",
                            self.unfolder.get_current_step())
                if self.unfolder.get_current_step() + 1 > max_user_step:
                    # Cosmetic break, avoid to do useless test
                    break

                ## TEST
                ## Make autoadjustment... of max_steps during the solving!
                ## Juste appeler squery_solve avec les memes parametres mais
                ## min_step augmenté de 1
                ## en restant inférieur au nb d'étapes demandées à l'origine
                ## => on bénéficie des shifts précédents et on évite de reconstruire le problème...

                # Hack: reuse the current Unfolder (without init with query)
                # => do not do this at home!
                self.unfolder._CLUnfolder__locked = False
                # Avoid a new test of satisfiability on the same current_step
                # (we already know that there is no more solution for it)
                # => increment __steps_before_check
                self.unfolder._CLUnfolder__steps_before_check += 1
                ret = self.unfolder.squery_is_satisfied(max_user_step)
                if not ret:
                    # The problem is definitly not satisfiable in the given
                    # number of steps
                    break
                # __steps_before_check can't be >= min_step
                # (see squery_is_satisfied() and squery_solve())
                min_step = self.unfolder.get_current_step()
                query.steps_before_check = min_step - 1

                print("testing AUTO CORRECT max step:", min_step)
                continue


            # Get the solution with the less number of activated states
            small_fsol = self.less_activated_solution(dimacs_front_sol)

            # Prune non essential places for the satisfiability of the property
            # => reduce the number of activated variables
            current_mac = self.__prune_frontier_solution(small_fsol, query, min_step)

            yield current_mac
            # mac_list.append(current_mac)

            # Keep a list of frontier values to be banned in the next search
            # - Get all activated frontiers on the current DimacsFrontierSol
            # - Build a list of their opposite values
            forbidden_frontier_values.append(
                [-var for var in current_mac.activated_frontier_values]
            )
            LOGGER.debug(
                "MCLA::mac_exhaustive_search: forbidden frontiers: %s",
                forbidden_frontier_values
            )

    def next_mac(self, query, max_step):
        """Search a Minimal Access Condition for the given query.

        This is a function very similar to __mac_exhaustive_search(), but it
        returns only 1 solution.
        Satisfiability tests and the banishment of previous solutions must be
        done before the call.

        ---

        On cherche ensuite d'abord des solutions non minimales (actuellement 2) via:
        self.__sq_dimacs_frontier_solutions(query, nb_step, 2)

        Pour ensuite les élaguer en supprimant les places non essentielles à la
        satisfiabilité de la propriété via:
        small_fsol = self.less_activated_solution(lfsol)
        current_mac = self.__prune_frontier_solution(small_sol, query, nb_step)

        Ce processus itératif est le plus "time-consuming" car nous n'avons pas
        le controle sur les solutions fournies par SAT et les solutions non minimales
        sont en général très éloignées de la solution minimale, c.-à-d. contiennent
        beaucoup plus de places (ces places excédentaires sont dispensables).

        :param max_step: Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        :type query: <MCLSimpleQuery>
        :type max_step: <int>
        :return: A FrontierSolution which is a wrapper of a symbolic
            representation of frontier values and timings from a raw solution.
        :rtype: <FrontierSolution> - <list <str>>
        """
        # Solutions differ on frontiers: Search only 2 different solutions
        # to avoid all activated solutions
        # Get <tuple <DimacsFrontierSol>>
        lfsol = self.__sq_dimacs_frontier_solutions(query, max_step, 2)
        if not lfsol:
            return

        # Get the solution with the less number of activated states
        small_fsol = self.less_activated_solution(lfsol)

        # Prune non essential places for the satisfiability of the property
        # => reduce the number of activated variables
        current_mac = self.__prune_frontier_solution(small_fsol, query, max_step)

        return FrontierSolution.from_dimacs_front_sol(current_mac)


    def mac_search(self, query, max_step):
        """Wrapper for __mac_exhaustive_search(); return a generator of
        FrontierSolution objects.

        ..note:: __mac_exhaustive_search() returns DimacsFrontierSols

        @param query: MCLSimpleQuery
        @param max_step: int - Number of steps allowed in the unfolding;
            the horizon on which the properties must be satisfied
        @return: <generator <FrontierSolution>>
        """
        ## TODO convert start_property in dim_start !! => eviter la recompilation des propriétés
        # Get <generator <DimacsFrontierSol>>
        mac_list = self.__mac_exhaustive_search(query, max_step)
        # Convert to a generator of FrontierSolution objects
        #return tuple(FrontierSolution.from_dimacs_front_sol(mac) for mac in mac_list)
        for mac in mac_list:
            yield FrontierSolution.from_dimacs_front_sol(mac)

    ## Inhibitors ##############################################################

    def is_strong_activator(self, act_sol, query):
        """Test if an activation condition is a strong one (independent of timing)

        ..refactor note: Unclear function

        final property becomes the invariant property
        => Ex: test if "C3 and C4" will never append if start property is "A1 and C1 and B1"
        Cf TestMCLAnalyser...

        @return: False if the problem is satisfiable, True otherwise
        @param act_sol: FrontierSolution with activation condition
        @param query: the query used for act_sol condition
        """
        max_step = len(act_sol.ic_sequence) - 1
        print("max steps:", max_step)

        # Same timings, force inactivated frontiers
        query_1 = MCLSimpleQuery.from_frontier_sol_same_timing(act_sol, self.unfolder)

        inv_prop = 'not(' + query.final_prop + ')'
        #(None, None, 'A1 and C1 and B1', ['', 'h2', '', ''])
        print("naive query1:", query_1.final_prop, query_1.inv_prop,
              query_1.start_prop, query_1.variant_prop)

        query_2 = MCLSimpleQuery(None, inv_prop, None)
        #(None, 'not(C3 and C4)', None, [])
        print("query 2:", query_2.final_prop, query_2.inv_prop,
              query_2.start_prop, query_2.variant_prop)
        query_1 = query_2.merge(query_1)
        #(None, 'not(C3 and C4)', 'A1 and C1 and B1', ['', 'h2', '', '']
        print("merged query1:", query_1.final_prop, query_1.inv_prop,
              query_1.start_prop, query_1.variant_prop)

        return not(self.sq_is_satisfiable(query_1, max_step))


    def next_inhibitor(self, mac, query):
        """
        ..refactor note: Unclear function

        if st_prop contains negation of mac conditions,
        return a different mac if any
        same parameters as __mac_exhaustive_search
        Used for search on cluster

        @param mac: FrontierSolution
        @param query: MCLSimpleQuery
        @param max_step: int - number of dynamical step
        @return: a list of variables (list<string>)
        """
        # query with mac init frontier places and  timing
        inh_query1 = MCLSimpleQuery.from_frontier_sol_same_timing(mac,
                                                                  self.unfolder)
        # query with negation of final prop as invariant property
        if not query.final_prop:
            raise MCLException("Query must have a final property")
        if not query.inv_prop:
            inv_prop = "not (" + query.final_prop + ")"
        else:
            inv_prop = query.inv_prop + "and (not (" + query.final_prop +"))"
        inh_query2 = MCLSimpleQuery(None, inv_prop, None)
        # init + timing + final property not reachable
        inh_query = inh_query1.merge(inh_query2)
        max_step = len(inh_query.variant_prop)
        assert(max_step == len(mac.ic_sequence))

        # search solutions - diseable aux clauses
        self.unfolder.set_include_aux_clauses(False)
        next_inhib = self.next_mac(inh_query, max_step)
        self.unfolder.set_include_aux_clauses(True)
        return next_inhib


    def mac_inhibitor_search(self, mac, query, max_sol):
        """Search inhibitors for a mac scenario

        ..refactor note: Unclear function

        @param mac: the mac
        @param query: the property enforced by the mac
        @param max_sol: limit on the number of solutions
        @param max_sol: maximum number of solutions

        @return: a list of frontier solution
        """
        # query with mac init frontier places and  timing
        inh_query1 = MCLSimpleQuery.from_frontier_sol(mac)
        # query with negation of final prop as invariant property
        if not query.final_prop:
            raise MCLException("Query must have a final property")
        if not query.inv_prop:
            inv_prop = "not (" + query.final_prop + ")"
        else:
            inv_prop = query.inv_prop + "and (not (" + query.final_prop +"))"
        inh_query2 = MCLSimpleQuery(None, inv_prop, None)
        # init + timing + final property not reachable
        inh_query = inh_query1.merge(inh_query2)

        max_step = len(inh_query.variant_prop)
        assert(max_step == len(mac.ic_sequence))
        # search solutions - diseable aux clauses
        self.unfolder.set_include_aux_clauses(False)
#        lsol = self.sq_frontier_solutions(inh_query, max_step, max_sol)
        lsol = tuple(self.mac_search(inh_query, max_step))
        self.unfolder.set_include_aux_clauses(True)
        return lsol

    def is_strong_inhibitor(self, in_sol, query):
        """Test if an activation condition inhibitor is a strong one
        (i.e independent of timing)

        ..refactor note: Unclear function

        @param in_sol: FrontierSolution with activation condition and inhibition
        @param query: the property which is inhibed
        """
        max_step = len(in_sol.ic_sequence) - 1

        # New timings, force inactivated frontiers
        query_1 = MCLSimpleQuery.from_frontier_sol_new_timing(in_sol, self.unfolder)
        # negation of the query
        if not query.final_prop:
            raise MCLException("Query must have a final property")
        if not query.inv_prop:
            inv_prop = "not (" + query.final_prop + ")"
        else:
            inv_prop = query.inv_prop + "and (not (" + query.final_prop +"))"
        inh_query = MCLSimpleQuery(None, inv_prop, None)
        query_1 = inh_query.merge(query_1)
        return not self.sq_is_satisfiable(query_1, max_step)




## Filename    : chart_simul_elem.py
## Author(s)   : Michel Le Borgne
## Created     : 04/2010
## Revision    : 
## Source      : 
##
## Copyright 2010 : IRISA/INRIA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and INRIA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall INRIA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## INRIA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA/INRIA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
## Contributor(s): Geoffroy Andrieux
##
"""
Building blocks for a simulator
"""

import sys

class Place(object):
    """
    A place represents a component of the state of the system
    """
    def __init__(self, g_state, name = None):
        """
        @param g_state: the component in the graphic model
        @param name: name of the place. mandatory if g_state is None
        """
        self.graph_state = g_state # place in chart_model (may be None)
        if self.graph_state:
            self.name = g_state.name
        else:
            self.name = name
        # permanent place (not unactivated by firing an outgoing transition)
        self.perm = False         
        self.activated = False    # state of the place
        self.out_transitions = [] # filled by transition constructor
        self.in_transitions = []  # idem
        self.ev_up = None         # event state up if used
        self.ev_down = None       # event state down if used
        
    def __str__(self):
        return self.name
    
    def is_place(self):
        """
        As it says
        """
        return True
    
    def is_input_event(self):
        """
        As it says
        """
        return False
    
    def is_input_place(self):
        """
        As it says
        """
        return False
    
    def set_perm(self):
        """
        Turn into a permanent node
        """
        self.perm = True
        
    def activate(self):
        """
        As it says
        """
        self.activated = True
        # for graphical representation of activity
        if self.graph_state:
            self.graph_state.activated = True
            self.graph_state.was_activated = True
        
    def desactivate(self):
        """
        places without graph representation are artificial inputs thus perm = True
        """
        if not self.perm:
            self.activated = False
            if self.graph_state:
                self.graph_state.activated = False # don't touch was_activated
            

    def force_desactivate(self):
        """
        Used to clear chart_model places activations
        desactivate all places - even permanent places
        """
        self.activated = False
        if self.graph_state:
            self.graph_state.activated = False    


class InputPlace(Place):
    """
    @param g_state: Corresponding place in CadbiomChart model (possibly None)
    @param name: Must be given if g_state=None
    An input place coming from the graph model is a sort of perm place
    An input place representing a free event has no corresponding element in chart model (g_state = None)
    """
    def __init__(self, g_state, name = None):
        if g_state:
            Place.__init__(self, g_state)
        else:
            Place.__init__(self, None, name)
        self.perm = True
        
    def is_input_place(self):
        return True
        
    def activate(self):
        self.activated = True
        
    def desactivate(self):
        return
                    
    def input_activate(self):
        """
        special activation for input
        """
        if self.activated and self.graph_state:
            self.graph_state.activated = True
        
    def input_desactivate(self):
        """
        special unsactivation for input
        """
        if self.graph_state:
            self.graph_state.activated = False
       
class STransition(object):
    """
    Represents a normal transition between  places - For optimization of induced transitions,
    STransitions have several origins and several targets. This is for taking care of 
    initializations when a macro state is activated and multiple un-activations for output
    transitions from places representing macros.
    """
    def __init__(self, ori, ext, gtrans=None):
        self.origin = ori       # Place
        self.target = ext       # Place
        self.g_trans = gtrans       # graph transition if any
        self.event = None       # an Event object
        self.condition = None   # condition object
        self.signal = None      # simple event object
        # register at origin as out transiton 
        self.origin.out_transitions.append(self)
        # similarly for in transitions
        self.target.in_transitions.append(self)  
        # simulation attributes
        # remember if target was activated by normal transition
        self.target_up = False  
        
    def __str__(self):
        return self.origin.graph_state.name+'->'+self.target.graph_state.name
    
    def set_event(self, evt):
        """
        standard setter
        """
        self.event = evt
        
    def set_condition(self, cond):
        """
        standard setter
        """
        self.condition = cond
        
    def set_signal(self, sig):
        """
        standard setter
        """
        self.signal = sig
                
    def activate(self):
        """
        activation
        """
        if self.g_trans:
            self.g_trans.activated = True
        if not self.target.activated:
            self.target_up = True
        self.target.activate()
        
    def desactivate(self):
        """
        unactivation
        """
        self.origin.desactivate()
            
    def init_desactivate(self):
        """
        initial RAZ
        """
        if self.g_trans:
            self.g_trans.activated = False
                
            
class Event(object):
    """
    An event is a clock from a Biosignal point of view - it is defined as a signal expression
    """
    def __init__(self, exp):
        self.expression = exp  # signal expression from compiler
        # if the expression is an identifier, actual value of the identifier
        self.activated = False 
        
    def is_place(self):
        """
        As it says
        """
        return False
    
    def is_input_event(self):
        """
        As it says
        """
        return False
    
    def set_activated(self, val):
        """
        @param val: activation value 
        """
        self.activated = val

class InputEvent(Event):
    """
    Special events for inputs
    """
    def __init__(self, exp):
        Event.__init__(self, exp)
        self.event_places = []
    
    def is_input_event(self):
        return True
    
    def add_ev_place(self, pla):
        """
        pla must be an input place
        """
        if pla.is_input_place():
            self.event_places.append(pla)
        else:
            print "add_ev_place: BUG"
            sys.exit(1)
        
    def set_activated(self, val):
        """
        overload the method of general class
        """
        self.activated = val
        if val:
            for pla in self.event_places:
                pla.input_activate()
        else:
            for pla in self.event_places:
                pla.input_desactivate()
            
                
class Condition(object):
    """
    A condition is a boolean expression on place values
    """
    def __init__(self, exp):
        self.expression = exp    
       



## Filename    : MCLTranslators.py
## Author(s)   : Michel Le Borgne
## Created     : 02/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
"""
Translators from guarded transition models to clause constraint models
"""
from cadbiom.models.biosignal.sig_expr import SigIdentExpr, SigSyncBinExpr, \
                            SigConstraintExpr

from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, \
                            compile_event, compile_constraint

from cadbiom.models.clause_constraints.CLDynSys import Literal, Clause


class MCLNoImplemented(Exception):
    """
    Exception for non implemented features
    """
    def __init__(self, mess):
        self.message = mess


class MCLSigExprVisitor(object):
    """
    visit a restricted SigExpression (type restrictions on default operands (both clocks)
    and when operands (clock, logical condition on states) and add clauses
    to logical dynamic system to encode y = f(X). SigExpressions are restricted.
    The variable y is represented as a literal for optimization reasons. It can be
    given as a parameter or generated as an intermediate literal.
    The generated clauses enforce the constraint y <=> f(X).
    All the expressions are memorized to implement common subexpression elimination.

    @param cldyn_sys: the dynamical system to be generated in clause form
    @param output: the variable name y or None for intermediate expressions)
    @param subexpressions: dictionary op#x#y -> literal for subexpressions memorization
    """
    def __init__(self, cldyn_sys, output, subexpressions = None):
        """
        @param cldyn_sys: clause constraint dynamical system to pupulate
        """
        self.reporter = cldyn_sys.report
        self.cldyn_sys = cldyn_sys
        # cannot test optimisation "turn on" on dictionary: empty dict => False
        if subexpressions == None:
            self.sub_dict = None
            self.opti = False
        else:
            self.sub_dict = subexpressions
            self.opti = True

        if output:
            # added because it may not appear in evolution
            self.cldyn_sys.add_var(output)
            self.output_lit = Literal(output, True) # output literal y
        else:
            self.output_lit = None

    def visit_sig_ident(self, sexpr):
        """
        if self.output(y) we have either y = x or (not y) = x - the y variable disappears
        either we are in an intermediate sexpr: returns the corresponding literal
        """
        # register variable names
        name = sexpr.name
        self.cldyn_sys.add_var(name)
        nlit = Literal(name, True)
        notnl = nlit.lit_not()
        # y = x : sexpr is top node of an expression
        if self.output_lit:
            if self.output_lit.sign:
                return nlit
            else:
                return notnl
        else:
            return nlit  # return the ident (Literal) if no output_var


    def visit_sig_not(self, sexpr):
        """
        if y is given and the expression is not f(X) we generate not y <=> f(X)
        """
        if self.output_lit: # y = not exp or (not y) = not expr
            slit = self.output_lit
            self.output_lit = self.output_lit.lit_not()
            notnl = sexpr.operand.accept(self)
            self.output_lit = slit # restore preceding value
        else :
            nlit = sexpr.operand.accept(self)
            notnl = nlit.lit_not()
        # don't generate more clauses than operand
        return notnl


    def visit_sig_sync(self, binsexpr):
        """
         the output_var is the next variable, the root of the tree
         we affect this name to the root and change output_var to None for the other BinExpression
        """
        # case y = f(X)
        if self.output_lit:
            nlit = self.output_lit
            notnl = nlit.lit_not()
        # sub expression: an auxiliary literal is introduced (nlit = f(X))
        else:
            name = self.cldyn_sys.add_aux_var()
            nlit = Literal(name, True)
            notnl = Literal(name, False)

        # recursive visits of operands
        olit = self.output_lit
        self.output_lit = None
        nlit1 = binsexpr.left_h.accept(self)
        notnl1 = nlit1.lit_not()
        nlit2 = binsexpr.right_h.accept(self)
        notnl2 = nlit2.lit_not()
        self.output_lit = olit

        # is a similar expression already visited?
        operator = binsexpr.operator
        if self.opti:
            var = search_sub_exp(operator, nlit1, nlit2, True, self.sub_dict)
            if var:
                if self.output_lit: # y = var
                    notnvar = var.lit_not()
                    cl1 = Clause([notnl, var])
                    cl2 = Clause([nlit, notnvar])
                    self.cldyn_sys.add_clause(cl1)
                    self.cldyn_sys.add_clause(cl2)
                    return self.output_lit
                return var

        # no sub expression: clauses generation for nlit = f(X)
        if operator == 'and': # nlit = lh and rh
            cl1 = Clause([nlit1, notnl])         # not nlit or not lh
            cl2 = Clause([nlit2, notnl])         # not nlit or not rh
            cl3 = Clause([notnl1, notnl2, nlit]) # nlit or not lh or not rh

        if operator == 'or':    # nlit =  or rh
            cl1 = Clause([notnl1, nlit])        # nlit or not lh
            cl2 = Clause([notnl2, nlit])        # nlit or not rh
            cl3 = Clause([nlit1, nlit2, notnl])   # not nlit or not lh or not rh

        self.cldyn_sys.add_clause(cl1)
        self.cldyn_sys.add_clause(cl2)
        self.cldyn_sys.add_clause(cl3)

        # sub expression memorization
        if self.opti:
            key = operator + '#' + nlit1.code_name() + '#' + nlit2.code_name()
            self.sub_dict[key] = nlit
        return nlit

    def visit_sig_const(self, exp):
        """
        compile a constant signal
        """
        if self.output_lit:
            nlit = self.output_lit
            notnl = nlit.lit_not()

            if exp.value:
                cla = nlit  # y = True
            else:
                cla = notnl  # y = False

        # the internal nodes will be affected with the name '_lit'
        else:
            name = self.cldyn_sys.add_aux_var()
            nlit = Literal(name, True)
            notnl = Literal(name, False)
        # clause generation
        if exp.value:
            cla = nlit  # x = True
        else:
            cla = nlit.lit_not()
        self.cldyn_sys.add_clause(Clause([cla]))
        return nlit

    def visit_sig_bot(self, bex):
        """
        The always absent clock
        """
        raise MCLNoImplemented("SigBotVisitor: Not yet implemented")

    def visit_sig_default(self, dex):
        """
        @param dex: default expression: lhs default rhs

        default is restricted to clocks
        """
        # case y = f(X)
        if self.output_lit:
            nlit = self.output_lit
            notnl = nlit.lit_not()
        # sub expression: an auxiliary literal is introduced (nlit = f(X))
        else:
            name = self.cldyn_sys.add_aux_var()
            nlit = Literal(name, True)
            notnl = Literal(name, False)

        # check if both operands are clocks
        if not dex.is_clock(self.cldyn_sys.symb_tab, check_clock):
            mess = "Default signal:"+dex.__str__()+" is not a clock"
            self.reporter.display(mess)
            return nlit # no meaning

        # visit recursively each operand
        olit = self.output_lit
        self.output_lit = None
        nl1 = dex.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = dex.right_h.accept(self)
        notnl2 = nl2.lit_not()
        self.output_lit = olit

        # is a similar expression already visited?
        operator = 'default'
        if self.opti:
            var = search_sub_exp(operator, nl1, nl2, True, self.sub_dict)
            if var:
                if self.output_lit: # y = var
                    notnvar = var.lit_not()
                    cl1 = Clause([notnl, var])
                    cl2 = Clause([nlit, notnvar])
                    self.cldyn_sys.add_clause(cl1)
                    self.cldyn_sys.add_clause(cl2)
                    return self.output_lit
                return var

        # generate nlit = nl1 or nl2
        self.cldyn_sys.add_clause(Clause([notnl, nl1, nl2]))  # nl => nl1 or nl2
        self.cldyn_sys.add_clause(Clause([notnl1, nlit]))     # nl1 => nlit
        self.cldyn_sys.add_clause(Clause([notnl2, nlit]))     # nl2 => nlit

        # sub expression memorization
        if self.opti:
            key = operator + '#' + nl1.code_name() + '#' + nl2.code_name()
            self.sub_dict[key] = nlit
        return nlit

    def visit_sig_when(self, wex):
        """
        When expression
        @param wex: when expression

        the left operand must be a clock and the right one must be
        a boolean exp on states and inputs (checked on syntax analysis)
        """
        # case y = f(X)
        if self.output_lit:
            nlit = self.output_lit
            notnl = nlit.lit_not()
        # sub expression: an auxiliary literal is introduced (nlit = f(X))
        else:
            name = self.cldyn_sys.add_aux_var()
            nlit = Literal(name, True)
            notnl = Literal(name, False)

        # check if left operand is clock
        if not wex.left_h.is_clock(self.cldyn_sys.symb_tab, check_clock):
            mess = "Default signal:"+wex.__str__()+" is not a clock"
            self.reporter.display(mess)
            return nlit # no meaning

        # visit recursively each operand
        olit = self.output_lit
        self.output_lit = None
        nl1 = wex.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = wex.right_h.accept(self)
        notnl2 = nl2.lit_not()
        self.output_lit = olit

        # is a similar expression already visited?
        operator = 'when'
        if self.opti:
            var = search_sub_exp(operator, nl1, nl2, False, self.sub_dict)
            if var:
                if self.output_lit: # y = var
                    notnvar = var.lit_not()
                    cl1 = Clause([notnl, var])
                    cl2 = Clause([nlit, notnvar])
                    self.cldyn_sys.add_clause(cl1)
                    self.cldyn_sys.add_clause(cl2)
                    return self.output_lit
                return var
        # generate nlit = nl1 and nl2
        #  nl1 and nl2 => nlit
        self.cldyn_sys.add_clause(Clause([notnl1, notnl2, nlit]))
        self.cldyn_sys.add_clause(Clause([notnl, nl1]))     # nlit => nl1
        self.cldyn_sys.add_clause(Clause([notnl, nl2]))     # nlit => nl2

        # sub expression memorization
        if self.opti:
            key = operator + '#' + nl1.code_name() + '#' + nl2.code_name()
            self.sub_dict[key] = nlit
        return nlit

    def visit_sig_constraint(self, cex):
        """
        generate clause constraints for clock constraints
        """
        if cex.constraint_name == SigConstraintExpr.SYNCHRO:
            gen_code_synchro(cex, self)
        elif cex.constraint_name == SigConstraintExpr.EXCLU:
            gen_code_exclu(cex, self)
        elif cex.constraint_name == SigConstraintExpr.INCL:
            gen_code_included(cex, self)
        else:
            self.reporter.display("Unknown constraint type: " + cex.name)

    def visit_sig_equa(self, eex):
        """
        TODO
        """
        raise MCLNoImplemented("MCLSigExprVisitor: Not yet implemented")

    def visit_sig_diff(self, dix):
        """
        TODO
        """
        raise MCLNoImplemented("MCLSigExprVisitor: Not yet implemented")

# helper fonctions
def gen_code_synchro(cex, visit):
    """
    code generator for a synchro constraint
    """
    if len(cex.arg) <2:
        visit.reporter.display("Incorrect clock synchro")
        return None
    exp = cex.arg[0]
    lit1 = exp.accept(visit)
    for exp in cex.arg[1:]:
        lit = exp.accept(visit)
        visit.cldyn_sys.add_clause(Clause([lit1.lit_not(), lit]))
        visit.cldyn_sys.add_clause(Clause([lit1, lit.lit_not()]))
        lit1 = lit

def gen_code_exclu(cex, visit):
    """
    code generator for an exclusion constraint
    """
    if len(cex.arg) <2:
        visit.reporter.display("Incorrect clock exclusion")
        return None
    llit = []
    for exp in cex.arg:
        lit = exp.accept(visit)
        llit.append(lit)
    for i in range(len(llit)):
        for j in range(i+1, len(llit)):
            visit.cldyn_sys.add_clause(Clause([llit[i].lit_not(),
                                               llit[j].lit_not()]))

def gen_code_included(cex, visit):
    """
    code generator for an inclusion constraint
    """
    if len(cex.arg) !=2:
        visit.reporter.display("Incorrect clock inclusion")
        return None
    lit1 = cex.arg[0].accept(visit)
    lit2 = cex.arg[1].accept(visit)
    visit.cldyn_sys.add_clause(Clause([lit1.lit_not(), lit2]))

def search_sub_exp(ope, op1, op2, com, sub_dict):
    """
    search if an expression a op b was already translated to clause constraints
    @param ope: operator (string)
    @param op1: left operand (literal)
    @param op2: right operand (literal)
    @param com: bool - true if operator is commutative
    @param sub_dict:directory of sub expressions
    """
    key = ope + '#' + op1.code_name() + '#' + op2.code_name()
    try:
        val_var = sub_dict[key]
        return val_var
    except KeyError:
        if com:
            key = ope + '#' + op2.code_name() + '#' + op1.code_name()
            try:
                val_var = sub_dict[key]
                return val_var
            except KeyError:
                return None
        else:
            return None

class GT2Clauses(object):
    """
    visit a charter_model based on guarded transitions and populate the associated
    CLSysDyn with clauses implementing the evolution constraints X' = f(X).
    Compute the frontier of the charter model. Frontier places are places without
    incoming transition or with a start incoming transition. In the later case,
    other transitions may exist.
    """

    def __init__(self, cldyn_sys, reporter, opti):
        """
        @param cldyn_sys: a new dynamic system in clause constraints.
        """
        self.cl_ds = cldyn_sys              # CLSysDyn to be populated
        self.reporter = reporter  # for error reporting
        if opti:
            self.sub_exp = dict()     # for subexpression elimination
        else:
            self.sub_exp = None


    def visit_chart_model(self, model):
        """
        Enter a model - after traversal, clock constraints are compiled
        """
        model.clean_code()
        model.get_root().accept(self)
        gen_aux_clock_constraints(self.cl_ds)
        gen_clock_constraints(model.constraints, self.cl_ds,
                              self.sub_exp, self.reporter)

    def visit_csimple_node(self, node):
        """
        simple node compilation
        """
        # register variable
        self.cl_ds.add_var(node.name)
        # frontier - registered for solution extraction
        if is_frontier(node):
            self.cl_ds.frontiers.append(node.name)
        else:
            self.cl_ds.no_frontiers.append(node.name)
        gen_simple_evolution(node, self.cl_ds, self.sub_exp, self.reporter)
        return

    def visit_cstart_node(self, node):
        """
        start node: nothing to do. Start nodes are frontier marks
        """
        pass

    def visit_ctrap_node(self, node):
        """
        trap node: compilation is done in transitions
        """
        pass

    def visit_cinput_node(self, node):
        """
        Input nodes generate free variables as free clocks
        We just have to register them for solution extraction
        """
        # register input variable
        self.cl_ds.add_input(node.name)

    def visit_cmacro_node(self, node):
        """
        Macro nodes are treated as simple nodes
        They must be empty!!
        """
        if len(node.sub_nodes) > 0:
            mess = "gt2Clause ==> ", "Macronodes not yet implemented!"
            self.reporter.display(mess)
        self.visit_csimple_node(node)

    def visit_cperm_node(self, node):
        """
        perm nodes are always activated - they do not appear in code
        """
        pass

    def visit_ctop_node(self, node):
        """
        visit all sub nodes
        """
        # sub states
        for sst in node.sub_nodes:
            sst.accept(self)

# helper functions

def check_clock(vva):
    """
    simple check
    """
    return vva[0] == 'clock'

def gen_transition_clock(trans, cl_ds, sub_exp, reporter):
    """
    @param trans: a transition
    @param cl_ds: logical (clauses) dynamic system to generate
    @param sub_exp: expression
    @param reporter: reporter for error reporting

    @return:  a literal representing the transition clock
    @warning: The clause constraint dynamic system cl_ds is modified
    """
    # condition compiling in SigExpr
    node_exp = SigIdentExpr(trans.ori.name)
    if trans.condition:
        cond_sexp = compile_cond(trans.condition, cl_ds.symb_tab, reporter)
        if trans.ori.is_perm():
            log_sexp = cond_sexp
        else:
            log_sexp = SigSyncBinExpr('and', node_exp, cond_sexp)
    else : # no condition on incoming transition
#        if trans.ori.is_perm():
#            log_sexp = None # stand for True!!
#        else:
        log_sexp = node_exp
    # condition compiling to clauses
    #temp_cond = cl_ds.add_aux_var()
    if log_sexp:
        sexpv = MCLSigExprVisitor(cl_ds, None, sub_exp)
        v_cond = log_sexp.accept(sexpv)
        notv_cond = v_cond.lit_not()

    # event compiling
    if trans.event:
        ev_sexpr, sex, free_clocks = compile_event(trans.event,
                                                   cl_ds.symb_tab,
                                                   True, reporter)
        for hne in free_clocks:
            cl_ds.add_free_clock(hne)
        if ev_sexpr.is_ident():
            h_event = Literal(trans.event, True)
        else:
            sexpv = MCLSigExprVisitor(cl_ds, None, sub_exp)
            h_event = ev_sexpr.accept(sexpv)

        if not log_sexp: # trans_h = ev because cond = True
            return h_event
        # register (trans.ori, h_event) since trans.ori is an
        # induction place of h_event
        if not trans.ori.is_perm():
            cl_ds.add_place_clock(trans.ori.name, h_event.name)
        nh_event = h_event.lit_not()
        nnn = cl_ds.add_aux_var() # name of the transition clock
        h_tr = Literal(nnn, True)
        noth_tr = Literal(nnn, False)
        cl1 = Clause([noth_tr, h_event])           # not tr_h or  ev
        cl_ds.add_clause(cl1)
        cl2 = Clause([noth_tr, v_cond])            #  not tr_h or  cond
        cl_ds.add_clause(cl2)
        # tr_h or (not cond) or (not ev)
        cl3 = Clause([nh_event, notv_cond, h_tr])
        cl_ds.add_clause(cl3)
        return h_tr

    else:
        return v_cond

def gen_transition_list_clock(ltr, cl_ds, sub_exp, reporter):
    """
    @param ltr: a list of transitions - must be non_empty
    @param cl_ds: logical (clauses) dynamic system to generate
    @param sub_exp: expression
    @param reporter: reporter for error reporting
    @return  a literal representing the default of the transition clocks or None if empty list
    The clause constraint dynamic system cl_ds is modified
    """
    if ltr == []:
         return None
    ltr_h = []
    for trans in ltr:
        if trans.ori.is_start():
            continue # trans is a pseudo transition
        if not trans.clock: # avoid multiple transition compilation
            tr_h = gen_transition_clock(trans, cl_ds, sub_exp, reporter)
            trans.clock = tr_h
        else:
            tr_h = trans.clock
        ltr_h.append(tr_h)
    if len(ltr_h) == 1:
        return ltr_h[0]
    h_name = cl_ds.add_aux_var() # name of the default clock
    h_def = Literal(h_name, True)
    nh_def = Literal(h_name, False)
    # add all the clauses h_def or not h_i
    for fcl in ltr_h:
        cl0 = Clause([h_def, fcl.lit_not()])
        cl_ds.add_clause(cl0)
    # add the clause (not h_def) or h1 or ...or hp
    ltr_h.append(nh_def)
    cl0 = Clause(ltr_h)
    cl_ds.add_clause(cl0)

    return h_def

def gen_simple_evolution(splace, cl_ds, sub_exp, reporter):
    """
    generate evolution equation of a place as clauses constraints
    @param splace: a simple place
    @param cl_ds: logical (clauses) dynamic system to generate
    @param sub_exp: expression
    @param reporter: reporter for error reporting
    @return  nothing
    The clause constraint dynamic system cl_ds is modified
    """
    pname = splace.name
    p_lit = Literal(pname, True)
    np_lit = Literal(pname, False)
    pnext = pname+"`"
    pnext_lit = Literal(pnext, True)
    npnext_lit = Literal(pnext, False)
    h_in = gen_transition_list_clock(splace.incoming_trans,
                                      cl_ds, sub_exp, reporter)
    h_out = gen_transition_list_clock(splace.outgoing_trans,
                                      cl_ds, sub_exp, reporter)
    # clauses for evolution equation
    if h_in and h_out:
        # not X' or h_in or X
        cla = Clause([npnext_lit, h_in, p_lit])
        cl_ds.add_clause(cla)
        # not X' or h_in or not h_out
        cla = Clause([npnext_lit, h_in, h_out.lit_not()])
        cl_ds.add_clause(cla)
        # X' or not h_in
        cla = Clause([pnext_lit, h_in.lit_not()])
        cl_ds.add_clause(cla)
        # X' or not X or h_out
        cla = Clause([pnext_lit, np_lit, h_out])
        cl_ds.add_clause(cla)

    elif h_in:
        # not X' or h_in or X
        cla = Clause([npnext_lit, h_in, p_lit])
        cl_ds.add_clause(cla)
        # X' or not h_in
        cla = Clause([pnext_lit, h_in.lit_not()])
        cl_ds.add_clause(cla)
        # X' or not X
        cla = Clause([pnext_lit, np_lit])
        cl_ds.add_clause(cla)

    elif h_out:
        # not X' or X
        cla = Clause([npnext_lit, p_lit])
        cl_ds.add_clause(cla)
        # not X' or not h_out
        cla = Clause([npnext_lit, h_out.lit_not()])
        cl_ds.add_clause(cla)
        # X' or not X or h_out
        cla = Clause([pnext_lit, np_lit, h_out])
        cl_ds.add_clause(cla)

    else:
        # not X' or X
        cla = Clause([npnext_lit, p_lit])
        cl_ds.add_clause(cla)
        # X' or not X
        cla = Clause([pnext_lit, np_lit])
        cl_ds.add_clause(cla)

def gen_aux_clock_constraints(cl_ds):
    """
    generate structural clock constraints:
    not h or A1 or A2 ... or Ap
    for all couples (A,h) where A is an induction place of h
    """
    # first constraint: a clock depends on its inductive places
    for hname in cl_ds.place_clocks.keys():
        hlit = Literal(hname, False)
        llit = [hlit]
        for pname in cl_ds.place_clocks[hname]:
            plit = Literal(pname, True)
            llit.append(plit)
        cl_ds.add_aux_clause(Clause(llit))

def gen_clock_constraints(constraints_txt, cl_ds, sub_exp, reporter):
    """
    cleaning clock constraints
    """
    text_const = ""
    for cha in constraints_txt:
        if  cha != ';' and cha != '\n':
            text_const += cha
        else:
            if cha == ';':
                gen_code_clock_const(text_const, cl_ds, sub_exp, reporter)
            text_const = ""


def gen_code_clock_const(const_txt, cl_ds, sub_exp, reporter):
    """
    code generator for clock constraints
    """
    ctexpr = compile_constraint(const_txt, cl_ds.symb_tab, reporter)
    ctexpv = MCLSigExprVisitor(cl_ds, None, sub_exp)
    ctexpr.accept(ctexpv)


def is_frontier(node):
    """
    Find if a node is a frontier node
    """
    itrans = node.incoming_trans
    if itrans == []:
        return True
    else:
        for trans in itrans:
            if trans.ori.is_start():
                return True
    return False

